import click

from corgie import scheduling, helpers, stack
from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords

from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
)
from corgie.spec import (
    spec_to_stack,
    spec_to_layer_dict_readonly,
    spec_to_layer_dict_overwrite,
)

from corgie.cli.compute_field import ComputeFieldJob
import json
from copy import deepcopy
import numpy as np


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--spec_path",
    nargs=1,
    type=str,
    required=True,
    help="JSON spec relating src stacks, src z to dst z",
)
@corgie_optgroup("Copy Method Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--blend_xy", nargs=1, type=int, default=0)
@corgie_option("--pad", nargs=1, type=int, default=512)
@corgie_option("--crop", nargs=1, type=int, default=None)
@corgie_option("--processor_spec", nargs=1, type=str, multiple=True, required=True)
@corgie_option("--processor_mip", nargs=1, type=int, multiple=True, required=True)
@click.option("--clear_nontissue_field", type=str, default=True)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@corgie_option("--suffix", nargs=1, type=str, default=None)
@click.pass_context
def compute_field_by_spec(
    ctx,
    spec_path,
    chunk_xy,
    blend_xy,
    pad,
    crop,
    processor_spec,
    processor_mip,
    clear_nontissue_field,
    start_coord,
    end_coord,
    coord_mip,
    suffix,
):

    scheduler = ctx.obj["scheduler"]
    if suffix is None:
        suffix = ""
    else:
        suffix = f"_{suffix}"

    with open(spec_path, "r") as f:
        spec = json.load(f)

    src_layers = spec_to_layer_dict_readonly(spec["src"])
    tgt_layers = spec_to_layer_dict_readonly(spec["tgt"])

    # if force_chunk_xy:
    #     force_chunk_xy = chunk_xy
    # else:
    #     force_chunk_xy = None

    # if force_chunk_z:
    #     force_chunk_z = chunk_z
    # else:
    #     force_chunk_z = None
    if crop is None:
        crop = pad

    reference_layer = src_layers[list(src_layers.keys())[0]]
    dst_layers = spec_to_layer_dict_overwrite(
        spec["dst"], reference_layer=reference_layer, default_type="field"
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    for dst_z in range(*bcube.z_range()):
        spec_z = str(dst_z)
        if spec_z in spec["job_specs"].keys():
            for job_spec in spec["job_specs"][spec_z]:
                src_stack = spec_to_stack(job_spec, "src", src_layers)
                tgt_stack = spec_to_stack(job_spec, "tgt", tgt_layers)
                dst_layer = dst_layers[str(job_spec["dst_img"])]
                ps = json.loads(processor_spec[0])
                ps["ApplyModel"]["params"]["val"] = job_spec["mask_id"]
                ps["ApplyModel"]["params"]["scale"] = job_spec["scale"]
                processor_spec = (json.dumps(ps),)
                job_bcube = bcube.reset_coords(
                    zs=job_spec["src_z"], ze=job_spec["src_z"] + 1, in_place=False
                )
                tgt_z_offset = job_spec["tgt_z"] - job_spec["src_z"]
                compute_field_job = ComputeFieldJob(
                    src_stack=src_stack,
                    tgt_stack=tgt_stack,
                    dst_layer=dst_layer,
                    chunk_xy=chunk_xy,
                    chunk_z=1,
                    blend_xy=blend_xy,
                    processor_spec=processor_spec,
                    pad=pad,
                    crop=crop,
                    bcube=job_bcube,
                    tgt_z_offset=tgt_z_offset,
                    suffix=suffix,
                    processor_mip=processor_mip,
                    clear_nontissue_field=clear_nontissue_field,
                )
                scheduler.register_job(
                    compute_field_job,
                    job_name="ComputeField {},{}".format(
                        job_bcube, job_spec["mask_id"]
                    ),
                )
    scheduler.execute_until_completion()



import click
import six
from copy import deepcopy

from corgie import scheduling, argparsers, helpers, stack

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords

from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)


class FillNearestJob(scheduling.Job):
    def __init__(
        self, src_stack, dst_stack, bcube, mip, radius, chunk_xy,
    ):
        """Write image, filling in masked regions with nearest non-masked image

        This is useful when a model requires a filled in image.
        """
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.bcube = bcube
        self.mip = mip
        self.radius = radius
        self.chunk_xy = chunk_xy
        super().__init__()

    def task_generator(self):
        chunks = self.dst_stack.get_layers()[0].break_bcube_into_chunks(
            bcube=self.bcube,
            chunk_xy=self.chunk_xy,
            chunk_z=1,
            mip=self.mip,
            return_generator=True,
        )

        tasks = (
            FillNearestTask(
                self.src_stack,
                self.dst_stack,
                mip=self.mip,
                bcube=chunk,
                radius=self.radius,
            )
            for chunk in chunks
        )
        corgie_logger.info(
            f"Yielding fill nearest tasks for bcube: {self.bcube}, MIP: {self.mip}"
        )

        yield tasks


class FillNearestTask(scheduling.Task):
    def __init__(
        self, src_stack, dst_stack, mip, bcube, radius=2,
    ):
        super().__init__(self)
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.mip = mip
        self.bcube = bcube
        self.radius = radius
        self.count_threshold = 10

    def get_masks(self, data_dict, bcube):
        agg_mask = helpers.zeros(
            (bcube.x_size(self.mip), bcube.y_size(self.mip)), dtype=bool
        )
        mask_layers = self.dst_stack.get_layers_of_type(["mask"])
        mask_layer_names = [l.name for l in mask_layers]
        for n, d in six.iteritems(data_dict):
            if n in mask_layer_names:
                if agg_mask is None:
                    agg_mask = d
                else:
                    agg_mask = (agg_mask + d) > 0
        return agg_mask

    def execute(self):
        radii = [i for r in range(1, self.radius + 1) for i in [r, -r]]
        z = self.bcube.z_range()[0]
        _, src_data_dict = self.src_stack.read_data_dict(
            bcube=self.bcube, mip=self.mip, add_prefix=False, translation_adjuster=None,
        )
        layer = self.dst_stack.get_layers_of_type("img")[0]
        agg_img = src_data_dict[f"{layer.name}"]
        agg_mask = self.get_masks(data_dict=src_data_dict, bcube=self.bcube)
        mask_count = agg_mask.sum()
        k = 0
        while (mask_count > self.count_threshold) and (k < len(radii)):
            corgie_logger.debug(f"mask_count={mask_count}")
            corgie_logger.debug(f"radius={radii[k]}")
            bcube = self.bcube.reset_coords(
                zs=z + radii[k], ze=z + radii[k] + 1, in_place=False
            )
            _, src_data_dict = self.src_stack.read_data_dict(
                bcube=bcube, mip=self.mip, add_prefix=False, translation_adjuster=None,
            )
            img = src_data_dict[f"{layer.name}"]
            agg_img[agg_mask] = img[agg_mask]
            mask = self.get_masks(data_dict=src_data_dict, bcube=bcube)
            agg_mask = (agg_mask == 1) * (mask == 1)
            mask_count = agg_mask.sum()
            k += 1
        layer.write(agg_img, bcube=self.bcube, mip=self.mip)


@click.command()
# Layers
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where aligned stack will go",
)
@corgie_option("--suffix", nargs=1, type=str, default=None)
@corgie_optgroup("Fill Nearest Method Specification")
@corgie_option("--mip", nargs=1, type=int, required=True)
@corgie_option("--radius", nargs=1, type=int, default=2)
@corgie_option("--chunk_xy", nargs=1, type=int, default=1024)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def fill_nearest(
    ctx,
    src_layer_spec,
    dst_folder,
    chunk_xy,
    start_coord,
    end_coord,
    coord_mip,
    suffix,
    mip,
    radius,
    force_chunk_z=1,
):
    scheduler = ctx.obj["scheduler"]

    if suffix is None:
        suffix = "_seethrough"
    else:
        suffix = f"_{suffix}"

    crop, pad = 0, 0
    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)
    src_stack.folder = dst_folder
    dst_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=dst_folder,
        name="dst",
        types=["img", "mask"],
        readonly=False,
        suffix=suffix,
        overwrite=True,
        force_chunk_z=force_chunk_z,
    )
    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    fill_nearest_job = FillNearestJob(
        src_stack=src_stack,
        dst_stack=dst_stack,
        bcube=bcube,
        radius=radius,
        mip=mip,
        chunk_xy=chunk_xy,
    )
    # create scheduler and execute the job
    scheduler.register_job(
        fill_nearest_job, job_name="Fill Nearest Block {}".format(bcube)
    )

    scheduler.execute_until_completion()
    result_report = (
        f"Rendered layers {[str(l) for l in src_stack.get_layers_of_type('img')]}. "
        f"Results in {[str(l) for l in dst_stack.get_layers_of_type('img')]}"
    )
    corgie_logger.info(result_report)




import click
import procspec

from corgie import scheduling, argparsers, helpers

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)

from corgie.cli.common import ChunkedJob
from corgie.cli.apply_processor import ApplyProcessorJob
import json


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--spec_path",
    nargs=1,
    type=str,
    required=True,
    help="JSON spec relating src stacks, src z to dst z",
)
@corgie_option(
    "--dst_layer_spec",
    nargs=1,
    type=str,
    required=True,
    help="Specification for the destination layer.",
)
@corgie_option("--reference_key", nargs=1, type=str, default="img")
@corgie_optgroup("Apply Processor Method Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_option("--blend_xy", nargs=1, type=int, default=0)
@corgie_option(
    "--pad",
    nargs=1,
    type=int,
    default=512,
)
@corgie_option("--crop", nargs=1, type=int, default=None)
@corgie_option("--processor_spec", nargs=1, type=str, multiple=True, required=True)
@corgie_option("--processor_mip", nargs=1, type=int, multiple=True, required=True)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def apply_processor_by_spec(
    ctx,
    src_layer_spec,
    spec_path,
    dst_layer_spec,
    processor_spec,
    pad,
    crop,
    chunk_xy,
    start_coord,
    processor_mip,
    end_coord,
    coord_mip,
    blend_xy,
    chunk_z,
    reference_key,
):
    scheduler = ctx.obj["scheduler"]

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    with open(spec_path, "r") as f:
        spec = set(json.load(f))

    reference_layer = None
    if reference_key in src_stack.layers:
        reference_layer = src_stack.layers[reference_key]

    dst_layer = create_layer_from_spec(
        dst_layer_spec,
        allowed_types=["img", "mask"],
        default_type="img",
        readonly=False,
        caller_name="dst_layer",
        reference=reference_layer,
        overwrite=True,
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    if crop is None:
        crop = pad

    for z in range(*bcube.z_range()):
        if z in spec:
            job_bcube = bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
            apply_processor_job = ApplyProcessorJob(
                src_stack=src_stack,
                dst_layer=dst_layer,
                chunk_xy=chunk_xy,
                chunk_z=chunk_z,
                blend_xy=blend_xy,
                processor_spec=processor_spec,
                pad=pad,
                crop=crop,
                bcube=job_bcube,
                processor_mip=processor_mip,
            )

            # create scheduler and execute the job
            scheduler.register_job(
                apply_processor_job, job_name="Apply Processor {}".format(job_bcube)
            )
    scheduler.execute_until_completion()



import copy

import torch

from corgie import constants, exceptions

from corgie.log import logger as corgie_logger
from corgie.boundingcube import BoundingCube
from corgie.layers.base import register_layer_type, BaseLayerType
from corgie import helpers


def get_extra_interpolate_parameters():
    # torch.nn.function.interpolate complains if this
    # argument is not provided, but it doesn't exist in older versions
    from packaging import version

    if version.parse(torch.__version__) <= version.parse("1.4.0"):
        return {}
    return {"recompute_scale_factor": False}


class VolumetricLayer(BaseLayerType):
    def __init__(self, data_mip=None, **kwargs):
        super().__init__(**kwargs)
        self.data_mip = data_mip

    def read(self, bcube, mip, **kwargs):
        indexed_bcube = self.indexing_scheme(bcube, mip, kwargs)
        if self.data_mip is not None:
            if mip <= self.data_mip:
                result_data_mip = super().read(
                    bcube=indexed_bcube, mip=self.data_mip, **kwargs
                )
                result = result_data_mip
                for _ in range(mip, self.data_mip):
                    result = self.get_upsampler()(result)
            elif mip > self.data_mip:
                # TODO: consider restricting MIP difference to prevent memory blow-up
                result_data_mip = super().read(
                    bcube=indexed_bcube, mip=self.data_mip, **kwargs
                )
                result = result_data_mip
                for _ in range(self.data_mip, mip):
                    result = self.get_downsampler()(result)
            else:
                # TODO: consider restricting MIP difference to prevent memory blow-up
                result_data_mip = super().read(
                    bcube=indexed_bcube, mip=self.data_mip, **kwargs
                )
                result = result_data_mip
                for _ in range(self.data_mip, mip):
                    result = self.get_downsampler()(result)
        else:
            result = super().read(bcube=indexed_bcube, mip=mip, **kwargs)

        return result

    def write(self, data_tens, bcube, mip, **kwargs):
        indexed_bcube = self.indexing_scheme(bcube, mip, kwargs)
        super().write(data_tens=data_tens, bcube=indexed_bcube, mip=mip, **kwargs)

    def indexing_scheme(self, bcube, mip, kwargs):
        return bcube

    def break_bcube_into_chunks(
        self,
        bcube,
        chunk_xy,
        chunk_z,
        mip,
        flatten=True,
        chunk_xy_step=None,
        chunk_z_step=None,
        **kwargs
    ):
        """Default breaking up of a bcube into smaller bcubes (chunks).
        Returns a list of chunks
        Args:
           bcube: BoundingBox for region to be broken into chunks
           chunk_size: tuple for dimensions of chunk that bbox will be broken into
           mip: int for MIP level at which chunk_xy is dspecified
        """
        indexed_bcube = self.indexing_scheme(bcube, mip, kwargs)

        x_range = indexed_bcube.x_range(mip=mip)
        y_range = indexed_bcube.y_range(mip=mip)
        z_range = indexed_bcube.z_range()

        if chunk_xy_step is None:
            chunk_xy_step = chunk_xy
        if chunk_z_step is None:
            chunk_z_step = chunk_z

        xy_chunks = []
        flat_chunks = []
        for zs in range(z_range[0], z_range[1], chunk_z_step):
            xy_chunks.append([])
            for xs in range(x_range[0], x_range[1], chunk_xy_step):
                xy_chunks[-1].append([])
                for ys in range(y_range[0], y_range[1], chunk_xy_step):
                    chunk = BoundingCube(
                        xs, xs + chunk_xy, ys, ys + chunk_xy, zs, zs + chunk_z, mip=mip
                    )

                    xy_chunks[-1][-1].append(chunk)
                    flat_chunks.append(chunk)

        if flatten:
            return flat_chunks
        else:
            return xy_chunks


@register_layer_type("img")
class ImgLayer(VolumetricLayer):
    def __init__(self, *args, num_channels=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_channels = num_channels

    def get_downsampler(self):
        def downsampler(data_tens):
            return torch.nn.functional.interpolate(
                data_tens.float(),
                mode="bilinear",
                scale_factor=1 / 2,
                align_corners=False,
                **get_extra_interpolate_parameters()
            )

        return downsampler

    def get_upsampler(self):
        def upsampler(data_tens):
            return torch.nn.functional.interpolate(
                data_tens.float(),
                mode="bilinear",
                scale_factor=2.0,
                align_corners=False,
                **get_extra_interpolate_parameters()
            )

        return upsampler

    def get_num_channels(self, *args, **kwargs):
        return self.num_channels

    def get_default_data_type(self):
        return "uint8"


@register_layer_type("segmentation")
class SegmentationLayer(VolumetricLayer):
    def __init__(self, *args, num_channels=1, **kwargs):
        if num_channels != 1:
            raise exceptions.ArgumentError(
                "Segmentation layer 'num_channels'",
                "Segmentation layer must have 1 channels. 'num_channels' provided: {}".format(
                    num_channels
                ),
            )
        super().__init__(*args, **kwargs)

    def read(self, dtype=None, **kwargs):
        return self.read_backend(transpose=False, **kwargs).squeeze()

    def get_downsampler(self):
        def downsampler(data_tens):
            downs_data = torch.nn.functional.interpolate(
                data_tens.float(),
                mode="nearest",
                scale_factor=1 / 2,
                align_corners=False,
                **get_extra_interpolate_parameters()
            )
            return downs_data

        return downsampler

    def get_upsampler(self):
        def upsampler(data_tens):
            ups_data = torch.nn.functional.interpolate(
                data_tens.float(),
                mode="nearest",
                scale_factor=2.0,
                align_corners=False,
                **get_extra_interpolate_parameters()
            )
            return ups_data

        return upsampler

    def get_num_channels(self, *args, **kwargs):
        return 1

    def get_default_data_type(self):
        return "uint64"


@register_layer_type("field")
class FieldLayer(VolumetricLayer):
    """Residuals are specified at a relative resolution based on MIP
    """

    def __init__(self, *args, num_channels=2, **kwargs):
        if num_channels != 2:
            raise exceptions.ArgumentError(
                "Field layer 'num_channels'",
                "Field layer must have 2 channels. 'num_channels' provided: {}".format(
                    num_channels
                ),
            )
        super().__init__(*args, **kwargs)

    def get_downsampler(self):
        def downsampler(data_tens):

            downs_data = torch.nn.functional.interpolate(
                data_tens.float(),
                mode="bilinear",
                scale_factor=1 / 2,
                align_corners=False,
                **get_extra_interpolate_parameters()
            )
            return downs_data * 0.5

        return downsampler

    def get_upsampler(self):
        def upsampler(data_tens):
            ups_data = torch.nn.functional.interpolate(
                data_tens.float(),
                mode="bilinear",
                scale_factor=2.0,
                align_corners=False,
                **get_extra_interpolate_parameters()
            )
            return ups_data * 2

        return upsampler

    def get_num_channels(self, *args, **kwargs):
        return 2

    def get_default_data_type(self):
        return "float32"


@register_layer_type("mask")
class MaskLayer(VolumetricLayer):
    def __init__(self, binarization=None, num_channels=1, **kwargs):
        self.binarizer = helpers.Binarizer(binarization)
        if num_channels != 1:
            raise exceptions.ArgumentError(
                "Mask layer 'num_channels'",
                "Mask layer must have 1 channels. 'num_channels' provided: {}".format(
                    num_channels
                ),
            )
        super().__init__(**kwargs)

    def read(self, **kwargs):
        data_tens = super().read(**kwargs)
        data_bin = self.binarizer(data_tens)
        return data_bin

    def get_downsampler(self):
        def downsampler(data_tens):
            return torch.nn.functional.max_pool2d(data_tens.float(), kernel_size=2)

        return downsampler

    def get_upsampler(self):
        def upsampler(data_tens):
            return torch.nn.functional.interpolate(
                data_tens.float(),
                mode="nearest",
                scale_factor=2.0,
                **get_extra_interpolate_parameters()
            )

        return upsampler

    def get_num_channels(self, *args, **kwargs):
        return 1

    def get_default_data_type(self):
        return "uint8"


@register_layer_type("section_value")
class SectionValueLayer(VolumetricLayer):
    def __init__(self, *args, num_channels=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_channels = num_channels

    # TODO: insert custom indexing here.

    def get_num_channels(self, *args, **kwargs):
        return 1

    def indexing_scheme(self, bcube, mip, kwargs):
        new_bcube = copy.deepcopy(bcube)
        if "channel_start" in kwargs and "channel_end" in kwargs:
            channel_start = kwargs["channel_start"]
            channel_end = kwargs["channel_end"]
            del kwargs["channel_start"], kwargs["channel_end"]
        else:
            channel_start = 0
            channel_end = self.num_channels

        new_bcube.reset_coords(channel_start, channel_end, 0, 1, mip=mip)
        return new_bcube

    def supports_voxel_offset(self):
        return False

    def supports_chunking(self):
        return False

    def get_default_data_type(self):
        return "float32"


@register_layer_type("fixed_field")
class FixedFieldLayer(FieldLayer):
    """Residuals are specified at a fixed resolution, regardless of MIP.
    For example, a field at MIP2 may have residuals specified in MIP0 
    pixels.

    NOTE: It is not recommended to create new fields based on this
    class. This class was intended to support existing fields created
    by legacy code. 
    """

    def __init__(self, *args, fixed_mip=0, **kwargs):
        super().__init__(*args, **kwargs)
        self.fixed_mip = fixed_mip

    def get_downsampler(self):
        def downsampler(data_tens):

            downs_data = torch.nn.functional.interpolate(
                data_tens.float(),
                mode="bilinear",
                scale_factor=1 / 2,
                align_corners=False,
                **get_extra_interpolate_parameters()
            )
            return downs_data

        return downsampler

    def get_upsampler(self):
        def upsampler(data_tens):
            ups_data = torch.nn.functional.interpolate(
                data_tens.float(),
                mode="bilinear",
                scale_factor=2.0,
                align_corners=False,
                **get_extra_interpolate_parameters()
            )
            return ups_data

        return upsampler

    def read(self, bcube, mip, **kwargs):
        return super().read(bcube, mip, **kwargs) / (2 ** (mip - self.fixed_mip))

    # def write(self, data_tens, bcube, mip, **kwargs):
    #     super().write(data_tens=data_tens*(2**self.fixed_mip),
    #                   bcube=indexed_bcube,
    #                   mip=mip,
    #                   **kwargs)


@register_layer_type("float_tensor")
class FloatTensorLayer(VolumetricLayer):
    """Generic 4D float tensors"""
    def __init__(self, *args, num_channels=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_channels = num_channels

    def get_downsampler(self):
        def downsampler(data_tens):
            return torch.nn.functional.interpolate(
                data_tens.float(),
                mode="bilinear",
                scale_factor=1 / 2,
                align_corners=False,
                **get_extra_interpolate_parameters()
            )

        return downsampler

    def get_upsampler(self):
        def upsampler(data_tens):
            return torch.nn.functional.interpolate(
                data_tens.float(),
                mode="bilinear",
                scale_factor=2.0,
                align_corners=False,
                **get_extra_interpolate_parameters()
            )

        return upsampler

    def get_num_channels(self, *args, **kwargs):
        return self.num_channels

    def get_default_data_type(self):
        return "float32"



import torch
import click
import json
from corgie.log import logger as corgie_logger
from corgie import scheduling, helpers, stack
from corgie.helpers import BoolFn
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)


class CombineMasksJob(scheduling.Job):
    def __init__(
        self,
        src_stack,
        exp,
        dst_layer,
        mip,
        bcube,
        pad,
        chunk_xy,
        chunk_z=1,
    ):
        """
        Combine masks according to boolean expression

        Args:
            layers ({key: Layer})
            exp (list of lists)
            output_field (Layer)
            mip (int)
            bcube (BoundingCube)
            pad (int)
            chunk_xy (int)
            chunk_z (int)
        """
        super().__init__()
        self.src_stack = src_stack
        self.exp = exp
        self.dst_layer = dst_layer
        self.mip = mip
        self.pad = pad
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z

    def task_generator(self):
        chunks = self.dst_layer.break_bcube_into_chunks(
            bcube=self.bcube, chunk_xy=self.chunk_xy, chunk_z=self.chunk_z, mip=self.mip
        )

        tasks = []
        for chunk in chunks:
            task = CombineMasksTask(
                src_stack=self.src_stack,
                exp=self.exp,
                dst_layer=self.dst_layer,
                mip=self.mip,
                bcube=chunk,
                pad=self.pad,
            )
            tasks.append(task)

        corgie_logger.debug(
            "Yielding CombineMasksTask for bcube: {}, MIP: {}".format(
                self.bcube, self.mip
            )
        )
        yield tasks


class CombineMasksTask(scheduling.Task):
    def __init__(
        self,
        src_stack,
        exp,
        dst_layer,
        mip,
        bcube,
        pad,
    ):
        """Evaluate a Boolean function of masks that may be from different layers and
        with different z_offsets. The Boolean function is expressed as a set of nested
        linear thresholded neurons, as described in helpers.BoolFn.

        Args:
            layers ({key: Layer}): dict with values of mask layers indexed with key that's
                referenced in boolean expression
            exp (BoolFn dict): BoolVar must have keys "weight", "key", and "offset", and must not
                have "inputs" as a key
                be lists of tuples with weight, layer key, and offset.
                For example:
                ```
                {
                "inputs": [
                    {
                    "inputs": [
                        { "weight": 1, "key": "a", "offset": 0 },
                        { "weight": -1, "key": "b", "offset": -1 },
                        { "weight": 1, "key": "c", "offset": 2 }
                    ],
                    "threshold": 2
                    },
                    {
                    "inputs": [
                        { "weight": 1, "key": "a", "offset": 1 },
                        { "weight": 1, "key": "c", "offset": 0 }
                    ],
                    "threshold": 1
                    }
                ],
                "threshold": 0
                }
                ```
                evaluates as
                `((a[0] - b[-1] + c[2] > 2) + (a[1] - c[0]) > 1) > 0`.
            output_field (Layer)
            mip (int)
            bcube (BoundingCube): xy location of where to sample all fields;
                z locations of where to write output
            pad (int)
        """
        super().__init__()
        self.src_stack = src_stack
        self.exp = exp
        self.dst_layer = dst_layer
        self.mip = mip
        self.pad = pad
        self.bcube = bcube

    def execute(self):
        corgie_logger.debug(f"CombineMaskTask, {self.bcube}")
        pbcube = self.bcube.uncrop(self.pad, self.mip)
        bf = BoolFn(self.exp)
        for z in range(*self.bcube.z_range()):

            def eval(exp):
                w, k, o = exp["weight"], exp["key"], exp["offset"]
                bcube = pbcube.reset_coords(zs=z + o, ze=z + o + 1, in_place=False)
                layer = self.src_stack.layers[k]
                return w * layer.read(bcube=bcube, mip=self.mip)

            m = bf(eval).to(torch.uint8)
            m = helpers.crop(m, self.pad)
            bcube = self.bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
            self.dst_layer.write(m, bcube=bcube, mip=self.mip)

        # This task can be used with caching
        for layer in self.src_stack.layers.values():
            layer.flush(self.mip)


@click.command()
# Layers
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_layer_spec",
    nargs=1,
    type=str,
    required=True,
    help="Specification for the destination layer. Must be an image or mask type.",
)
@corgie_optgroup("Combine Mask Method Specification")
@corgie_option(
    "--exp",
    nargs=1,
    type=str,
    required=True,
    help="Boolean expression to be evaluated. "
    + "Must be a JSON-parseable string describing a nested set of linear threshold (LT) neurons. "
    + "LT neurons are defined as dict with keys 'inputs' and 'threshold'. "
    + "The 'threshold' is any float. The 'inputs' may be any LT neuron, as well as a BoolVar. "
    + "This specific BoolVar must be a dict that does not contain 'inputs' as a key, and "
    + "must contain the following keys: "
    + "1. 'weight' as float for coefficient of the mask; e.g. as boolean literal, -1 can be considered NOT. "
    + "2. 'key' as str for layer name in src_stack "
    + "2. 'offset' as int for positive offset from the current z index ",
)
@corgie_option("--mip", nargs=1, type=int, required=True)
@corgie_option("--pad", nargs=1, type=int, default=0)
@corgie_option("--force_chunk_xy", nargs=1, type=int)
@corgie_option("--force_chunk_z", nargs=1, type=int, default=1)
@corgie_option("--chunk_xy", nargs=1, type=int, default=1024)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def combine_masks(
    ctx,
    src_layer_spec,
    dst_layer_spec,
    exp,
    chunk_xy,
    chunk_z,
    force_chunk_xy,
    force_chunk_z,
    start_coord,
    end_coord,
    coord_mip,
    mip,
    pad,
):
    scheduler = ctx.obj["scheduler"]

    if not force_chunk_xy:
        force_chunk_xy = chunk_xy

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)
    reference_layer = src_stack.reference_layer

    dst_layer = create_layer_from_spec(
        dst_layer_spec,
        allowed_types=["mask"],
        default_type="mask",
        readonly=False,
        caller_name="dst_layer",
        reference=reference_layer,
        force_chunk_xy=force_chunk_xy,
        force_chunk_z=force_chunk_z,
        overwrite=True,
    )
    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    combine_masks_job = CombineMasksJob(
        src_stack=src_stack,
        exp=json.loads(exp),
        dst_layer=dst_layer,
        mip=mip,
        bcube=bcube,
        pad=pad,
        chunk_xy=chunk_xy,
        chunk_z=chunk_z,
    )
    # create scheduler and execute the job
    scheduler.register_job(combine_masks_job, job_name="Combine Masks {}".format(bcube))

    scheduler.execute_until_completion()
    result_report = f"Results in {str(dst_layer)}"
    corgie_logger.info(result_report)



from corgie.stack import Stack
from corgie.log import logger as corgie_logger
from corgie.argparsers import create_layer_from_dict, create_layer_from_spec
import json

def spec_to_stack(spec, prefix, layers):
    """Create Stack by filtering a dict of layers

    Args:
        spec (dict): job_specs including src_img, src_mask, etc.
        prefix (str): src/tgt/dst
        layers (dict): int-indexed dict of layers

    job_spec will include index of layer to be used for job
    """
    stack = Stack()
    for suffix in ['img', 'mask', 'field']:
        spec_key = '{}_{}'.format(prefix, suffix)
        if spec_key in spec.keys():
            layer_id = str(spec[spec_key])
            layer = layers[layer_id]
            layer.name = suffix
            stack.add_layer(layer)
    return stack

def spec_to_layer_dict_readonly(layer_specs):
    """Create dict of layers from a corgie spec indexed by unique id

    These layers will be readonly

    Args:
        layer_specs (dict): layer specs indexed by unique id
        spec_key (str): src/tgt
    """
    layers = {}
    for k, s in layer_specs.items():
        corgie_logger.info(f'Creating layer no. {k}')
        layers[k] = create_layer_from_dict(s, readonly=True)
    return layers

def spec_to_layer_dict_overwrite(layer_specs, reference_layer, default_type):
    """Create dict of layers from a corgie spec indexed by unique id

    These layers will be of type overwrite

    Args:
        layer_specs (dict): layer specs indexed by unique id
        reference_layer (layer)
        default_type (str): e.g. img, field
    """
    layers = {}
    for k, s in layer_specs.items():
        layers[k] = create_layer_from_spec(json.dumps(s),
                                        default_type=default_type,
                                        readonly=False, 
                                        reference=reference_layer, 
                                        overwrite=True)
    return layers



import json
import math

import click
import torch
from corgie import helpers, residuals, scheduling
from corgie.argparsers import (
    corgie_optgroup,
    corgie_option,
    create_layer_from_dict,
)
from corgie.boundingcube import get_bcube_from_coords
from corgie.log import logger as corgie_logger
from corgie.spec import spec_to_layer_dict_readonly
from corgie.stack import FieldSet


class MergeRenderJob(scheduling.Job):
    def __init__(self, src_layers, src_specs, dst_layer, mip, pad, bcube, chunk_xy):
        """Render multiple images to the same destination image

        Args:
            src_layers ({'img':{Layers},
            src_specs (json): list of dicts with img, field, mask, z, & mask_id per island
                ranked by layer priority (first image overwrites later images)
            dst_layer (Stack)
            mip (int)
            pad (int)
            bcube (BoundingCube)
        """
        self.src_layers = src_layers
        self.src_specs = src_specs
        self.dst_layer = dst_layer
        self.mip = mip
        self.pad = pad
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        super().__init__()

    def task_generator(self):
        chunks = self.dst_layer.break_bcube_into_chunks(
            bcube=self.bcube,
            chunk_xy=self.chunk_xy,
            chunk_z=1,
            mip=self.mip,
            return_generator=True,
        )

        if "src_img" in self.src_specs[0]:
            tasks = (
                MergeRenderImageTask(
                    src_layers=self.src_layers,
                    src_specs=self.src_specs,
                    dst_layer=self.dst_layer,
                    mip=self.mip,
                    pad=self.pad,
                    bcube=input_chunk,
                )
                for input_chunk in chunks
            )
        else:
            tasks = (
                MergeRenderMaskTask(
                    src_layers=self.src_layers,
                    src_specs=self.src_specs,
                    dst_layer=self.dst_layer,
                    mip=self.mip,
                    pad=self.pad,
                    bcube=input_chunk,
                )
                for input_chunk in chunks
            )
        corgie_logger.info(
            f"Yielding render tasks for bcube: {self.bcube}, MIP: {self.mip}"
        )

        yield tasks


class MergeRenderImageTask(scheduling.Task):
    def __init__(self, src_layers, src_specs, dst_layer, mip, pad, bcube):
        """Render multiple images to the same destination image

        Args:
            src_layers ({Layer})
            src_specs (json): list of dicts with img, field, mask, z, & mask_id per island
                ranked by layer priority (first image overwrites later images)
            dst_layer (Stack)
            mip (int)
            pad (int)
            bcube (BoundingCube)
        """
        super().__init__(self)
        self.src_layers = src_layers
        self.src_specs = src_specs
        self.dst_layer = dst_layer
        self.mip = mip
        self.pad = pad
        self.bcube = bcube

    def execute(self):
        device = "cuda" if torch.cuda.is_available() else "cpu"

        # Field padding
        padded_bcube = self.bcube.uncrop(self.pad, self.mip)
        for k, specs in enumerate(self.src_specs[::-1]):
            src_z = specs["src_z"]
            dst_z = self.bcube.z_range()[0]

            corgie_logger.info(f"Load fields for {padded_bcube}")
            # backwards compatible
            if not isinstance(specs["src_field"], list):
                specs["src_field"] = [specs["src_field"]]
            mask_layer = self.src_layers[str(specs["src_mask"])]

            field_ids = list(map(str, specs["src_field"]))
            corgie_logger.info(f"field ids={field_ids}")
            z_list = specs.get("src_field_z", [src_z] * len(field_ids))
            fields = FieldSet([self.src_layers[n] for n in field_ids])
            field = fields.read(
                bcube=padded_bcube, z_list=z_list, mip=self.mip, device=device
            )
            bcube = padded_bcube.reset_coords(zs=src_z, ze=src_z + 1, in_place=False)

            # Extend image/mask cutout to account for field spread
            render_pad = int((field.max_vector() - field.min_vector()).max().ceil().tensor().item())
            snap_factor = 2 ** (max(self.mip, mask_layer.data_mip) - self.mip)
            render_pad = math.ceil(render_pad / snap_factor) * snap_factor
            render_pad = min(render_pad, 4096)  # Safety

            render_bcube = bcube.uncrop(render_pad, self.mip)
            corgie_logger.debug(f"render_pad: {render_pad}")

            # Move image/mask cutout to account for field drift
            img_trans = helpers.percentile_trans_adjuster(field)
            mask_trans = img_trans.round_to_mip(self.mip, mask_layer.data_mip)
            corgie_logger.debug(f"img_trans: {img_trans} | mask_trans: {mask_trans}")

            img_bcube = render_bcube.translate(
                x_offset=img_trans.y, y_offset=img_trans.x, mip=self.mip
            )
            mask_bcube = render_bcube.translate(
                x_offset=mask_trans.y, y_offset=mask_trans.x, mip=self.mip
            )

            if render_pad > 0:
                field = torch.nn.functional.pad(field, [render_pad, render_pad, render_pad, render_pad], mode='replicate')

            corgie_logger.info(f"Load masks for {mask_bcube}")
            mask_id = specs["mask_id"]
            mask_layer.binarizer = helpers.Binarizer(["eq", mask_id])
            mask = mask_layer.read(bcube=mask_bcube, mip=self.mip, device=device)
            mask = residuals.res_warp_img(
                mask.float(), field - mask_trans.to_tensor(device=field.device)
            ).tensor()
            mask = (mask > 0.4).bool()
            cropped_mask = helpers.crop(mask, self.pad + render_pad)

            corgie_logger.info(f"Load image for {img_bcube}")
            if cropped_mask.sum() == 0:
                cropped_img = torch.zeros_like(cropped_mask, dtype=torch.float)
            else:
                img_layer = self.src_layers[str(specs["src_img"])]
                img = img_layer.read(bcube=img_bcube, mip=self.mip, device=device)
                img = residuals.res_warp_img(
                    img.float(), field - img_trans.to_tensor(device=field.device)
                )
                cropped_img = helpers.crop(img, self.pad + render_pad)

            # write to composite image
            if k == 0:
                dst_img = cropped_img
                dst_img[~cropped_mask] = 0
            else:
                dst_img[cropped_mask] = cropped_img[cropped_mask]

        self.dst_layer.write(dst_img.cpu(), bcube=self.bcube, mip=self.mip)


class MergeRenderMaskTask(MergeRenderImageTask):
    """Render multiple masks into the same destination mask, with option to relabel

    Spec format:
        {
            'src_mask' (int)
            'src_field' (int)
            'mask_id' (int): the id of this mask in the source
            'relabel_id' (int): the id of this mask in the destination (optional)
            'src_z' (int)
        }
    """

    def execute(self):
        device = "cuda" if torch.cuda.is_available() else "cpu"
        padded_bcube = self.bcube.uncrop(self.pad, self.mip)
        for k, specs in enumerate(self.src_specs[::-1]):
            src_z = specs["src_z"]
            dst_z = self.bcube.z_range()[0]

            corgie_logger.info(f"Load fields for {padded_bcube}")
            # backwards compatible
            if not isinstance(specs["src_field"], list):
                specs["src_field"] = [specs["src_field"]]
            mask_layer = self.src_layers[str(specs["src_mask"])]

            field_ids = list(map(str, specs["src_field"]))
            corgie_logger.info(f"field ids={field_ids}")
            z_list = specs.get("src_field_z", [src_z] * len(field_ids))
            fields = FieldSet([self.src_layers[n] for n in field_ids])
            field = fields.read(
                bcube=padded_bcube, z_list=z_list, mip=self.mip, device=device
            )
            bcube = padded_bcube.reset_coords(zs=src_z, ze=src_z + 1, in_place=False)

            mask_trans = helpers.percentile_trans_adjuster(field)
            mask_trans = mask_trans.round_to_mip(self.mip, mask_layer.data_mip)
            corgie_logger.debug(f"mask_trans: {mask_trans}")

            mask_bcube = bcube.translate(
                x_offset=mask_trans.y, y_offset=mask_trans.x, mip=self.mip
            )

            corgie_logger.info(f"Load masks for {mask_bcube}")
            mask_id = specs["mask_id"]
            mask_layer.binarizer = helpers.Binarizer(["eq", mask_id])
            mask = mask_layer.read(bcube=mask_bcube, mip=self.mip, device=device)
            mask = residuals.res_warp_img(
                mask.float(), field - mask_trans.to_tensor(device=field.device)
            ).tensor()
            mask = (mask > 0.4).bool()
            cropped_mask = helpers.crop(mask, self.pad)

            relabel_id = torch.as_tensor(specs.get("relabel_id", k + 1), dtype=torch.uint8)
            if k == 0:
                dst_img = cropped_mask * relabel_id
                dst_img[~cropped_mask] = 0
            else:
                dst_img[cropped_mask] = cropped_mask[cropped_mask] * relabel_id

        self.dst_layer.write(dst_img.cpu(), bcube=self.bcube, mip=self.mip)


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--spec_path",
    nargs=1,
    type=str,
    required=True,
    help="JSON spec relating src stacks, src z to dst z",
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=False,
    help="Destination folder for the copied stack",
)
@corgie_optgroup("Render Method Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--force_chunk_xy", nargs=1, type=int)
@corgie_option("--pad", nargs=1, type=int, default=512)
@corgie_option("--mip", nargs=1, type=int, required=True)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@corgie_option("--suffix", nargs=1, type=str, default=None)
@click.pass_context
def merge_render(
    ctx,
    spec_path,
    dst_folder,
    chunk_xy,
    pad,
    start_coord,
    end_coord,
    coord_mip,
    force_chunk_xy,
    mip,
    suffix,
):

    scheduler = ctx.obj["scheduler"]
    if suffix is None:
        suffix = ""
    else:
        suffix = f"_{suffix}"

    corgie_logger.debug("Setting up layers...")
    # create layers
    # collect image paths
    # collect mask paths

    if not force_chunk_xy:
        force_chunk_xy = chunk_xy

    with open(spec_path, "r") as f:
        spec = json.load(f)

    src_layers = spec_to_layer_dict_readonly(spec["src"])
    reference_layer = src_layers[list(src_layers.keys())[0]]
    dst_layer = create_layer_from_dict(
        {"path": dst_folder, "type": "img"},
        reference=reference_layer,
        force_chunk_xy=force_chunk_xy,
        overwrite=True
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    for z in range(*bcube.z_range()):
        tgt_z = str(z)
        if tgt_z in spec["job_specs"].keys():
            job_bcube = bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
            render_job = MergeRenderJob(
                src_layers=src_layers,
                src_specs=spec["job_specs"][tgt_z],
                dst_layer=dst_layer,
                mip=mip,
                pad=pad,
                bcube=job_bcube,
                chunk_xy=chunk_xy,
            )
            scheduler.register_job(
                render_job, job_name="MergeRender {}".format(job_bcube)
            )
    scheduler.execute_until_completion()



import click
import procspec

from corgie import scheduling, argparsers, helpers

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)


class ComputeFieldJob(scheduling.Job):
    def __init__(
        self,
        src_stack,
        tgt_stack,
        dst_layer,
        chunk_xy,
        chunk_z,
        processor_spec,
        pad,
        crop,
        bcube,
        tgt_z_offset,
        processor_mip,
        suffix="",
    ):

        self.src_stack = src_stack
        self.tgt_stack = tgt_stack
        self.dst_layer = dst_layer
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.pad = pad
        self.crop = crop
        self.bcube = bcube
        self.tgt_z_offset = tgt_z_offset
        self.suffix = suffix  # in case this job wants to make more layers

        self.processor_spec = processor_spec
        self.processor_mip = processor_mip
        if isinstance(self.processor_spec, str):
            self.processor_spec = [self.processor_spec]
        if isinstance(self.processor_mip, int):
            self.processor_mip = [self.processor_mip]

        if len(self.processor_mip) != len(self.processor_spec):
            raise exceptions.CorgieException(
                "The number of processors doesn't "
                "match the number of specified processor MIPs"
            )

        super().__init__()

    def task_generator(self):
        self.dst_layer.declare_write_region(
            self.bcube, mips=self.processor_mip, chunk_xy=chunk_xy, chunk_z=chunk_z
        )

        all_layers = self.src_stack.get_layers() + self.tgt_stack.get_layers()

        last_mip = None
        for i in range(len(self.processor_spec)):
            this_proc = self.processor_spec[i]
            this_proc_mip = self.processor_mip[i]
            is_last_proc = i == len(self.processor_spec) - 1

            chunks = self.dst_layer.break_bcube_into_chunks(
                bcube=self.bcube,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                mip=this_proc_mip,
            )

            tasks = [
                ComputeFieldTask(
                    src_stack=self.src_stack,
                    tgt_stack=self.tgt_stack,
                    dst_layer=self.dst_layer,
                    processor_spec=this_proc,
                    mip=this_proc_mip,
                    pad=self.pad,
                    crop=self.crop,
                    tgt_z_offset=self.tgt_z_offset,
                    bcube=chunk,
                )
                for chunk in chunks
            ]

            corgie_logger.debug(
                "Yielding CF tasks for bcube: {}, MIP: {}".format(
                    self.bcube, this_proc_mip
                )
            )
            yield tasks

            if not is_last_proc:
                yield scheduling.wait_until_done
                next_proc_mip = self.processor_mip[i + 1]
                if this_proc_mip > next_proc_mip:
                    downsample_job = DownsampleJob(
                        src_layer=self.dst_layer,
                        chunk_xy=self.chunk_xy,
                        chunk_z=self.chunk_z,
                        mip_start=this_proc_mip,
                        mip_end=next_proc_mip,
                        bcube=self.bcube,
                    )
                    yield from downsample_job.task_generator
                    yield scheduling.wait_until_done

        if self.processor_mip[0] > self.processor_mip[-1]:
            # good manners
            # prepare the ground for the next you
            downsample_job = DownsampleJob(
                src_layer=self.dst_layer,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                mip_start=self.processor_mip[-1],
                mip_end=self.processor_mip[0],
                bcube=self.bcube,
            )


class ComputeFieldTask(scheduling.Task):
    def __init__(
        self,
        processor_spec,
        src_stack,
        tgt_stack,
        dst_layer,
        mip,
        pad,
        crop,
        tgt_z_offset,
        bcube,
    ):
        super().__init__()
        self.processor_spec = processor_spec
        self.src_stack = src_stack
        self.tgt_stack = tgt_stack
        self.dst_layer = dst_layer
        self.mip = mip
        self.pad = pad
        self.crop = crop
        self.tgt_z_offset = tgt_z_offset
        self.bcube = bcube

    def execute(self):
        src_bcube = self.bcube.uncrop(self.pad, self.mip)
        tgt_bcube = src_bcube.translate(z=self.tgt_z_offset)

        processor = procspec.parse_proc(
            spec_str=self.processor_spec, default_output="src_cf_field"
        )

        src_translation, src_data_dict = self.src_stack.read_data_dict(
            src_bcube, mip=self.mip, stack_name="src"
        )
        tgt_translation, tgt_data_dict = self.tgt_stack.read_data_dict(
            tgt_bcube, mip=self.mip, stack_name="tgt"
        )

        processor_input = {**src_data_dict, **tgt_data_dict}

        predicted_field = processor(processor_input)
        cropped_field = helpers.crop(predicted_field, self.crop)
        self.dst_layer.write(cropped_field, bcube=self.bcube, mip=self.mip)


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
#
@corgie_option(
    "--tgt_layer_spec",
    "-t",
    nargs=1,
    type=str,
    required=False,
    multiple=True,
    help="Target layer spec. Use multiple times to include all masks, fields, images. "
    "DEFAULT: Same as source layers",
)
@corgie_option(
    "--dst_layer_spec",
    nargs=1,
    type=str,
    required=False,
    help="Specification for the destination layer. Must be a field type."
    + " DEFAULT: source reference key path + /field/cf_field + (_{suffix})?",
)
@corgie_option("--reference_key", nargs=1, type=str, default="img")
@corgie_optgroup("Compute Field Method Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_option(
    "--pad",
    nargs=1,
    type=int,
    default=512,
)
@corgie_optgroup("")
@corgie_option("--crop", nargs=1, type=int, default=None)
@corgie_option("--processor_spec", nargs=1, type=str, multiple=True, required=True)
@corgie_option("--mip", nargs=1, type=int, multiple=True, required=True)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@corgie_option("--tgt_z_offset", nargs=1, type=str, default=1)
@click.option("--suffix", nargs=1, type=str, default=None)
@click.pass_context
def compute_field(
    ctx,
    src_layer_spec,
    tgt_layer_spec,
    dst_layer_spec,
    suffix,
    processor_spec,
    pad,
    crop,
    chunk_xy,
    start_coord,
    mip,
    end_coord,
    coord_mip,
    tgt_z_offset,
    chunk_z,
    reference_key,
):

    scheduler = ctx.obj["scheduler"]

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    tgt_stack = create_stack_from_spec(
        tgt_layer_spec, name="tgt", readonly=True, reference=src_stack
    )

    reference_layer = None
    if reference_key in src_stack.layers:
        reference_layer = src_stack.layers[reference_key]
    dst_layer = create_layer_from_spec(
        dst_layer_spec,
        allowed_types=["field"],
        default_type="field",
        readonly=False,
        caller_name="dst_layer",
        reference=reference_layer,
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    if crop is None:
        crop = pad

    compute_field_job = ComputeFieldJob(
        src_stack=src_stack,
        tgt_stack=tgt_stack,
        dst_layer=dst_layer,
        chunk_xy=chunk_xy,
        chunk_z=chunk_z,
        processor_spec=processor_spec,
        pad=pad,
        crop=crop,
        bcube=bcube,
        tgt_z_offset=tgt_z_offset,
        suffix=suffix,
        mip=mip,
    )

    # create scheduler and execute the job
    scheduler.register_job(
        align_block_job,
        job_name="Compute field {}, tgt z offset {}".format(bcube, tgt_z_offset),
    )
    scheduler.execute_until_completion()



import click
from cloudfiles import CloudFiles
from cloudvolume import PrecomputedSkeleton, Skeleton
import numpy as np
import os
import pickle

from corgie import scheduling, argparsers, stack
from corgie.log import logger as corgie_logger
from corgie.boundingcube import get_bcube_from_vertices
from corgie.argparsers import (
    LAYER_HELP_STR,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)
from cloudvolume.lib import Bbox


def get_skeleton(src_path, skeleton_id_str):
    cf = CloudFiles(src_path)
    return Skeleton.from_precomputed(cf.get(skeleton_id_str))


def get_skeleton_vert_neighbor_ids(sk, vert_id):
    vert_edges = np.where(
        np.isin(sk.edges[:, 0], vert_id) + np.isin(sk.edges[:, 1], vert_id)
    )[0]

    neighbors = sk.edges[vert_edges[-1]].flatten()

    # filter out self
    self_index = np.where(neighbors == vert_id)
    neighbors = np.delete(neighbors, self_index)

    return neighbors


def rip_out_verts(sk, vert_ids):
    vert_edges = np.where(
        np.isin(sk.edges[:, 0], vert_ids) + np.isin(sk.edges[:, 1], vert_ids)
    )[0]
    sk.edges = np.delete(sk.edges, vert_edges, axis=0)

    return sk.consolidate()


class FilterSkeletonsJob(scheduling.Job):
    def __init__(
        self,
        src_path,
        dst_path,
        skeleton_ids,
        bad_sections,
        z_start,
        z_end,
        skeleton_length_file=None,
    ):
        self.src_path = src_path
        self.dst_path = dst_path
        self.bad_sections = bad_sections
        self.z_start = z_start
        self.z_end = z_end

        if skeleton_ids is None:
            self.skeleton_ids = self.get_all_skeleton_ids()
        else:
            self.skeleton_ids = skeleton_ids
        self.skeleton_length_file = skeleton_length_file
        super().__init__()

    def task_generator(self):
        skeletons = self.get_skeletons(self.src_path)
        if self.z_start is not None and self.z_end is not None:
            bbox = Bbox((0, 0, self.z_start * 40), (10e8, 10e8, self.z_end * 40))
        else:
            bbox = None

        lengths = []
        for skeleton_id_str, sk in skeletons.items():
            deleted = 1
            if bbox is not None:
                sk = sk.crop(bbox)
            while deleted != 0:
                deleted = 0
                verts = sk.vertices
                vert_zs = verts[:, 2].copy()
                vert_zs /= 40  # hack
                vert_zs = vert_zs.astype(np.int32)

                bad_verts = np.where(np.isin(vert_zs, self.bad_sections))[0]

                # Find the next set of vertices to remove so that
                # we don't mess with indices, remove the in the next pass
                deleted = 0
                for bv in bad_verts:
                    neighbors = get_skeleton_vert_neighbor_ids(sk, bv)

                    # filter out other bad vertices
                    bad_index = np.where(np.isin(neighbors, bad_verts))
                    neighbors = np.delete(neighbors, bad_index)

                    assert np.isin(neighbors.flatten(), bad_verts).sum() == 0

                    # if there's a good neighbor, reasign all the endpoints to it
                    # else wait until the next round
                    if len(neighbors) != 0:
                        replacement = neighbors[0]
                        assert replacement not in bad_verts
                        ed = np.expand_dims(sk.edges, -1)
                        ed[np.where(ed == bv)] = replacement
                        sk.edges = ed.squeeze(-1)
                        # this leaves self-edges
                        deleted += 1
                # This removes self edges
                sk = sk.consolidate()
                new_v_count = sk.vertices.shape[0]

            verts = sk.vertices
            vert_zs = verts[:, 2].copy()
            vert_zs /= 40  # hack
            vert_zs = vert_zs.astype(np.int32)
            bad_verts = np.where(np.isin(vert_zs, self.bad_sections))[0]

            lengths.append((skeleton_id_str, sk.cable_length()))
            cf = CloudFiles(self.dst_path)
            cf.put(
                path=skeleton_id_str,
                content=sk.to_precomputed(),
                compress="gzip",
            )
        for n, l in sorted(lengths):
            print(l)
        import sys

        sys.exit(1)
        # print (len(bad_verts))
        # print (sk.vertices.shape)

    def get_skeletons(self, folder):
        skeleton_filenames = [str(skeleton_id) for skeleton_id in self.skeleton_ids]
        cf = CloudFiles(folder)
        skeleton_files = cf.get(skeleton_filenames)
        skeletons = {}
        for skeleton_file in skeleton_files:
            skeleton_id_str = skeleton_file["path"]
            skeleton = Skeleton.from_precomputed(skeleton_file["content"])
            skeletons[skeleton_id_str] = skeleton
        return skeletons

    def get_all_skeleton_ids(self):
        cf = CloudFiles(self.src_path)
        skeleton_filenames = cf.list(flat=True)
        skeleton_ids = []
        for skeleton_filename in skeleton_filenames:
            if ":" in skeleton_filename:
                # Fragment
                continue
            skeleton_ids.append(int(skeleton_filename))
        return skeleton_ids


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where skeletons are stored",
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where to store new skeletons",
)
@corgie_optgroup("Misc Parameters")
@corgie_option(
    "--bad_sections", "-b", multiple=True, type=int, help="Bad sections to filter out"
)
@corgie_option(
    "--bad_sections", "-b", multiple=True, type=int, help="Bad sections to filter out"
)
@corgie_option("--z_start", type=int, default=None)
@corgie_option("--z_end", type=int, default=None)
@corgie_option("--ids", multiple=True, type=int, help="Ids to transform")
@corgie_option("--ids_filepath", type=str, help="File containing ids to transform")
@click.pass_context
def filter_skeletons(
    ctx, src_folder, dst_folder, ids, bad_sections, ids_filepath, z_start, z_end
):
    scheduler = ctx.obj["scheduler"]

    corgie_logger.debug("Setting up layers...")

    skeleton_ids = ids
    if ids_filepath is not None:
        skeleton_ids = []
        with open(ids_filepath, "r") as f:
            line = f.readline()
            while line:
                skeleton_ids.append(int(line))
                line = f.readline()

    if len(skeleton_ids) == 0:
        skeleton_ids = None
    else:
        skeleton_ids = list(skeleton_ids)

    transform_skeletons_job = FilterSkeletonsJob(
        src_path=src_folder,
        dst_path=dst_folder,
        skeleton_ids=skeleton_ids,
        bad_sections=bad_sections,
        z_start=z_start,
        z_end=z_end,
    )

    scheduler.register_job(
        transform_skeletons_job,
        job_name="Filtering skeletons in {}".format(src_folder),
    )

    scheduler.execute_until_completion()
    result_report = f"Filtered skeletons stored at {dst_folder}. "
    corgie_logger.info(result_report)



import os
import copy
import numpy as np

from cloudfiles import CloudFiles

from corgie import layers
from corgie import exceptions
from corgie.log import logger as corgie_logger

from corgie.data_backends.base import DataBackendBase, BaseLayerBackend, \
        register_backend, str_to_backend


@register_backend("json")
class JSONDataBackend(DataBackendBase):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)
json_backend = str_to_backend("json")

class JSONLayerBase(BaseLayerBackend):
    """A directory with one text file per section
    """
    def __init__(self, path, backend, reference=None, overwrite=True, **kwargs):
        super().__init__(**kwargs)
        self.path = path
        self.dtype = 'O'
        self.backend = backend
        self.cf = CloudFiles(self.path, progress=False)

    def __str__(self):
        return "JSON {}".format(self.path)

    def get_sublayer(self, name, layer_type=None, path=None, **kwargs):
        if path is None:
            path = os.path.join(self.path, layer_type, name)

        if layer_type is None:
            layer_type = self.get_layer_type()

        return self.backend.create_layer(path=path, reference=self,
                layer_type=layer_type, **kwargs)

    def get_filename(self, z):
        return f'{z:06d}'

    def read_backend(self, bcube, **kwargs):
        z_range = bcube.z_range()
        corgie_logger.debug(f'Read from {str(self)}, z: {z_range}')
        data = []
        for z in z_range:
            f = self.cf.get_json(self.get_filename(z))
            data.append(f)
        return data

    def write_backend(self, data, bcube, **kwargs):
        z_range = range(*bcube.z_range())
        assert(len(data) == len(z_range))
        corgie_logger.debug(f'Write to {str(self)}, z: {z_range}')
        filepaths = [self.get_filename(z) for z in z_range]
        self.cf.put_jsons(zip(filepaths, data), cache_control='no-cache')

@json_backend.register_layer_type_backend("section_value")
class JSONSectionValueLayer(JSONLayerBase, layers.SectionValueLayer):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)



import click

from corgie import exceptions, stack, helpers, scheduling, argparsers
import json

from corgie.log import logger as corgie_logger
from corgie.boundingcube import get_bcube_from_coords

from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_dict,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)

from corgie.cli.compare_sections import CompareSectionsJob
from corgie.cli.combine_masks import CombineMasksJob


class DetectSlipMisalignmentsJob(scheduling.Job):
    def __init__(self, src_stack, dst_layer, mip, bcube, pad, chunk_xy, chunk_z):
        super().__init__()
        self.src_stack = src_stack
        self.dst_layer = dst_layer
        self.mip = mip
        self.bcube = bcube
        self.pad = pad
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z

    @classmethod
    def get_exp(cls):
        return {
            "inputs": [
                {"weight": -1, "key": -3, "offset": 0},
                {"weight": -1, "key": -2, "offset": 0},
                {"weight": -1, "key": -1, "offset": 0},
                {"weight": -1, "key": -1, "offset": 1},
                {"weight": -1, "key": -2, "offset": 2},
                {"weight": -1, "key": -3, "offset": 3},
            ],
            "threshold": -2,
        }

    def task_generator(self):
        exp = DetectSlipMisalignmentsJob.get_exp()
        combine_masks_job = CombineMasksJob(
            src_stack=self.src_stack,
            exp=exp,
            dst_layer=self.dst_layer,
            mip=self.mip,
            bcube=self.bcube,
            pad=self.pad,
            chunk_xy=self.chunk_xy,
            chunk_z=self.chunk_z,
        )
        yield from combine_masks_job.task_generator


class DetectStepMisalignmentsJob(scheduling.Job):
    def __init__(self, src_stack, dst_layer, mip, bcube, pad, chunk_xy, chunk_z):
        super().__init__()
        self.src_stack = src_stack
        self.dst_layer = dst_layer
        self.mip = mip
        self.bcube = bcube
        self.pad = pad
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z

    @classmethod
    def get_exp(cls):
        return {
            "inputs": [
                {
                    "inputs": [
                        {"weight": 1, "key": -3, "offset": 0},
                        {"weight": 1, "key": -2, "offset": 0},
                        {"weight": 1, "key": -1, "offset": 0},
                    ],
                    "threshold": 1,
                },
                {
                    "inputs": [
                        {"weight": -1, "key": -1, "offset": 1},
                        {"weight": -1, "key": -2, "offset": 2},
                        {"weight": -1, "key": -3, "offset": 3},
                    ],
                    "threshold": -1,
                },
            ],
            "threshold": 1,
        }

    def task_generator(self):
        exp = DetectStepMisalignmentsJob.get_exp()
        combine_masks_job = CombineMasksJob(
            src_stack=self.src_stack,
            exp=exp,
            dst_layer=self.dst_layer,
            mip=self.mip,
            bcube=self.bcube,
            pad=self.pad,
            chunk_xy=self.chunk_xy,
            chunk_z=self.chunk_z,
        )
        yield from combine_masks_job.task_generator


class DetectConsecutiveMasksJob(scheduling.Job):
    def __init__(
        self,
        src_stack,
        dst_layer,
        mip,
        bcube,
        pad,
        chunk_xy,
        chunk_z,
        num_consecutive=3,
        key="slip",
    ):
        super().__init__()
        self.src_stack = src_stack
        self.dst_layer = dst_layer
        self.num_consecutive = num_consecutive
        self.mip = mip
        self.bcube = bcube
        self.pad = pad
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.key = key

    @classmethod
    def get_exp(cls, n, key="slip"):
        """Return boolean expression for n consecutive sections.

        e.g. n=3 would return:
        {
            "inputs": [
                {
                    "inputs": [
                        {"weight": 1, "key": 0, "offset": -2},
                        {"weight": 1, "key": 0, "offset": -1},
                        {"weight": 1, "key": 0, "offset": 0},
                    ],
                    "threshold": 3,
                },
                {
                    "inputs": [
                        {"weight": 1, "key": 0, "offset": -1},
                        {"weight": 1, "key": 0, "offset": 0},
                        {"weight": 1, "key": 0, "offset": 1},
                    ],
                    "threshold": 3,
                },
                {
                    "inputs": [
                        {"weight": 1, "key": 0, "offset": 0},
                        {"weight": 1, "key": 0, "offset": 1},
                        {"weight": 1, "key": 0, "offset": 2},
                    ],
                    "threshold": 3,
                },
            ],
            "threshold": 1,
        }
        """
        o = {"inputs": [], "threshold": 0}
        for i in range(n):
            oi = {"inputs": [], "threshold": n - 1}
            for j in range(n):
                oij = {"weight": 1, "key": key, "offset": i + j - n + 1}
                oi["inputs"].append(oij)
            o["inputs"].append(oi)
        return o

    def task_generator(self):
        exp = DetectConsecutiveMasksJob.get_exp(n=self.num_consecutive, key=self.key)
        combine_masks_job = CombineMasksJob(
            src_stack=self.src_stack,
            exp=exp,
            dst_layer=self.dst_layer,
            mip=self.mip,
            bcube=self.bcube,
            pad=self.pad,
            chunk_xy=self.chunk_xy,
            chunk_z=self.chunk_z,
        )
        yield from combine_masks_job.task_generator


@click.command()
# Layers
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where aligned stack will go",
)
@corgie_option("--suffix", nargs=1, type=str, default=None)
@corgie_optgroup("Create Segmentation Masks Specification")
@corgie_option("--processor_spec", nargs=1, type=str, required=True, multiple=False)
@corgie_option(
    "--processor_mip",
    "-m",
    nargs=1,
    type=int,
    required=True,
    help="MIP of the input for creating masks",
)
@corgie_option(
    "--dst_mip",
    "-m",
    nargs=1,
    type=int,
    required=True,
    help="MIP of the final masks",
)
@corgie_option(
    "--compute_similarities/--skip_similarities",
    default=True,
    help="If not computing similarities, they must be defined as src_layer_specs. "
    + "Define similarities as masks with similar set to True, and False otherwise. "
    + "Label similarities with their z offset (e.g. -3, -2, -1).",
)
@corgie_option(
    "--similarity_threshold",
    nargs=1,
    type=float,
    default=0.1,
    help="If computing_similarities=True, similarity values above this threshold will be considered similar.",
)
@corgie_option(
    "--compute_slip_mask/--skip_slip_mask",
    default=True,
    help="If not computing slip mask, must be defined as src_layer_specs. "
    + "The mask should be named 'slip', and be True where there is a misalignment.",
)
@corgie_option(
    "--compute_step_mask/--skip_step_mask",
    default=True,
    help="If not computing step mask, it must be defined as src_layer_specs. "
    + "The mask should be named 'stpe', and be True where there is a misalignment.",
)
@corgie_option("--compute_affinity_mask/--skip_affinity_mask", default=True)
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_option("--force_chunk_xy", nargs=1, type=int, default=None)
@corgie_option("--pad", nargs=1, type=int, default=256)
@corgie_option("--crop", nargs=1, type=int, default=None)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def create_segmentation_masks(
    ctx,
    src_layer_spec,
    dst_folder,
    processor_spec,
    pad,
    crop,
    processor_mip,
    dst_mip,
    chunk_xy,
    chunk_z,
    force_chunk_xy,
    start_coord,
    end_coord,
    coord_mip,
    suffix,
    similarity_threshold,
    compute_similarities,
    compute_slip_mask,
    compute_step_mask,
    compute_affinity_mask,
):

    scheduler = ctx.obj["scheduler"]

    if suffix is None:
        suffix = "_aligned"
    else:
        suffix = f"_{suffix}"

    if crop is None:
        crop = pad

    corgie_logger.debug("Setting up layers...")

    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)
    src_stack.folder = dst_folder

    if force_chunk_xy is None:
        force_chunk_xy = chunk_xy

    dst_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=dst_folder,
        name="dst",
        types=["img", "mask"],
        readonly=False,
        suffix=suffix,
        force_chunk_xy=force_chunk_xy,
        overwrite=True,
    )

    corgie_logger.debug("Done!")

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    if compute_similarities:
        z_offsets = range(-1, -4, -1)
        for tgt_z_offset in z_offsets:
            if tgt_z_offset not in src_stack:
                dst_layer = dst_stack.create_sublayer(
                    name=tgt_z_offset,
                    layer_type="img",
                    overwrite=True,
                    layer_args={"dtype": "uint8"},
                )

                proc_spec = json.loads(processor_spec)
                if isinstance(proc_spec, dict):
                    assert str(tgt_z_offset) in proc_spec
                    proc_spec = json.dumps(proc_spec[str(tgt_z_offset)])
                else:
                    proc_spec = processor_spec

                compare_job = CompareSectionsJob(
                    src_stack=src_stack,
                    tgt_stack=src_stack,
                    dst_layer=dst_layer,
                    chunk_xy=chunk_xy,
                    processor_spec=proc_spec,
                    pad=pad,
                    bcube=bcube,
                    tgt_z_offset=tgt_z_offset,
                    suffix=suffix,
                    mip=processor_mip,
                    dst_mip=dst_mip,
                )
                scheduler.register_job(
                    compare_job,
                    job_name="Compare Sections Job {}, tgt z offset {}".format(
                        bcube, tgt_z_offset
                    ),
                )
        scheduler.execute_until_completion()
        result_report = f"Similarity results in {[str(l) for l in dst_stack.get_layers_of_type('img')]}"
        corgie_logger.info(result_report)

        # If similarities were just computed, add them to the src_stack as masks
        # Otherwise, they need to be included with src_stack_specs as masks
        # See arg help above for similarity mask and misalignment mask definitions.
        for layer_name in z_offsets:
            img_layer = dst_stack[layer_name]
            binarizer = {
                "binarization": ["gt", similarity_threshold],
                # "cv_params": {"cache": True},
            }
            layer_dict = {
                "path": img_layer.path,
                "name": img_layer.name,
                "type": "mask",
                "args": binarizer,
            }
            mask_layer = create_layer_from_dict(layer_dict, reference=dst_stack)
            src_stack.add_layer(mask_layer)

    corgie_logger.info("Computing slip & step masks")
    if compute_slip_mask:
        slip_layer = dst_stack.create_sublayer(
            name="slip", layer_type="mask", overwrite=True
        )
        slip_bcube = bcube.reset_coords(
            zs=bcube.z[0] + 1, ze=bcube.z[1] - 1, in_place=False
        )
        slip_misalignments_job = DetectSlipMisalignmentsJob(
            src_stack=src_stack,
            dst_layer=slip_layer,
            mip=dst_mip,
            bcube=slip_bcube,
            pad=pad,
            chunk_xy=chunk_xy,
            chunk_z=chunk_z,
        )
        scheduler.register_job(
            slip_misalignments_job,
            job_name="Detect Slip Misalignments {}".format(bcube),
        )
    if compute_step_mask:
        step_layer = dst_stack.create_sublayer(
            name="step", layer_type="mask", overwrite=True
        )
        step_bcube = bcube.reset_coords(
            zs=bcube.z[0] + 2, ze=bcube.z[1] - 2, in_place=False
        )
        step_misalignments_job = DetectStepMisalignmentsJob(
            src_stack=src_stack,
            dst_layer=step_layer,
            mip=dst_mip,
            bcube=step_bcube,
            pad=pad,
            chunk_xy=chunk_xy,
            chunk_z=chunk_z,
        )
        scheduler.register_job(
            step_misalignments_job,
            job_name="Detect Step Misalignments {}".format(bcube),
        )
    if compute_slip_mask or compute_step_mask:
        # Execute slip & step masks at the same time
        scheduler.execute_until_completion()
        result_report = f"Slip & step masks in in {str(slip_layer), str(step_layer)}"
        corgie_logger.info(result_report)

    if compute_affinity_mask:
        corgie_logger.info("Creating affinity masks")
        affinity_layer = dst_stack.create_sublayer(
            name="affinity", layer_type="mask", overwrite=True
        )
        three_consecutive_exp = DetectConsecutiveMasksJob.get_exp(n=3, key="slip")
        exp = {
            "inputs": [
                three_consecutive_exp,
                {"weight": 1, "key": "step", "offset": 0},
            ],
            "threshold": 0,
        }
        affinity_masks_job = CombineMasksJob(
            src_stack=dst_stack,
            exp=exp,
            dst_layer=affinity_layer,
            mip=dst_mip,
            bcube=bcube,
            pad=pad,
            chunk_xy=chunk_xy,
            chunk_z=chunk_z,
        )
        scheduler.register_job(
            affinity_masks_job, job_name="Affinity Masks {}".format(bcube)
        )
        scheduler.execute_until_completion()
        result_report = f"Results in {str(affinity_layer)}"
        corgie_logger.info(result_report)


import click
import torch
from math import log
from corgie.log import logger as corgie_logger
from corgie.boundingcube import get_bcube_from_coords
from corgie import scheduling, stack
from corgie.argparsers import (
    LAYER_HELP_STR,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)


class VoteJob(scheduling.Job):
    def __init__(
        self,
        input_fields,
        output_field,
        chunk_xy,
        bcube,
        z_offsets,
        mip,
        consensus_threshold=3.,
        blur_sigma=15.0,
        kernel_size=32,
        weights_layer=None,
    ):
        self.input_fields = input_fields
        self.output_field = output_field
        self.chunk_xy = chunk_xy
        self.bcube = bcube
        self.z_offsets = z_offsets
        self.mip = mip
        self.consensus_threshold = consensus_threshold
        self.blur_sigma = blur_sigma
        self.kernel_size = kernel_size
        self.weights_layer = weights_layer
        super().__init__()

    def task_generator(self):
        chunks = self.output_field.break_bcube_into_chunks(
            bcube=self.bcube, chunk_xy=self.chunk_xy, chunk_z=1, mip=self.mip
        )

        tasks = [
            VoteTask(
                input_fields=self.input_fields,
                output_field=self.output_field,
                mip=self.mip,
                bcube=chunk,
                z_offsets=self.z_offsets,
                consensus_threshold=self.consensus_threshold,
                blur_sigma=self.blur_sigma,
                kernel_size=self.kernel_size,
                weights_layer=self.weights_layer,
            )
            for chunk in chunks
        ]

        corgie_logger.debug(
            "Yielding VoteTask for bcube: {}, MIP: {}".format(self.bcube, self.mip)
        )
        yield tasks


class VoteTask(scheduling.Task):
    def __init__(
        self,
        input_fields,
        output_field,
        mip,
        bcube,
        z_offsets,
        consensus_threshold=3.,
        blur_sigma=15,
        kernel_size=32,
        weights_layer=None,
    ):
        """Find median-like field with highest priority from a set of fields

        Notes:
            Does not ignore identity fields.
            Padding & cropping is not necessary.

        Args:
            input_fields ([Layer]): match length of z_offsets
            output_field (Layer)
            mip (int)
            bcube (BoundingCube)
            z_offsets (range, [int]): offsets from z where fields should be accessed
            consensus_threshold (float): distance from winning subset that will still be 
                considered consensus
            blur_sigma (float): std of Gaussian that will blur weights of the fields
            weight_layer (Layer): if set, where intermediary weights will be written
        """
        super().__init__()
        self.input_fields = input_fields
        self.output_field = output_field
        self.mip = mip
        self.bcube = bcube
        self.z_offsets = z_offsets
        # match length of input_fields & z_offsets
        if len(self.input_fields) < len(self.z_offsets):
            self.input_fields = [self.input_fields[0]] * len(self.z_offsets)
        elif len(self.z_offsets) < len(self.input_fields):
            self.z_offsets = [self.z_offsets[0]] * len(self.input_fields)
        self.consensus_threshold = consensus_threshold
        self.blur_sigma = blur_sigma
        self.kernel_size = kernel_size
        self.weights_layer = weights_layer

    def execute(self):
        fields = []
        priorities = []
        for priority, (z_offset, field) in enumerate(zip(self.z_offsets, self.input_fields)):
            z = self.bcube.z_range()[0]
            bcube = self.bcube.reset_coords(
                zs=z + z_offset, ze=z + z_offset + 1, in_place=False
            )
            fields.append(field.read(bcube, mip=self.mip))
            priorities.append(torch.full_like(fields[-1][:,0,...], fill_value=len(self.input_fields) -priority))
        fields = torch.cat([f for f in fields]).field()
        priorities = torch.cat(priorities)
        weights = fields.get_priority_vote_weights(priorities=priorities, consensus_threshold=self.consensus_threshold)
        if self.weights_layer:
            self.weights_layer.write(data_tens=weights, bcube=self.bcube, mip=self.mip)
        voted_field = fields.smoothed_combination(weights=weights, blur_sigma=self.blur_sigma, kernel_size=self.kernel_size)
        self.output_field.write(data_tens=voted_field, bcube=self.bcube, mip=self.mip)


@click.command()
# Layers
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where aligned stack will go",
)
@corgie_option("--suffix", nargs=1, type=str, default=None)
@corgie_optgroup("Voting Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--z_offsets", multiple=True, type=int, default=[0])
@corgie_option("--consensus_threshold", nargs=1, type=float, default=3.)
@corgie_option("--blur_sigma", nargs=1, type=float, default=15.0)
@corgie_option("--kernel_size", nargs=1, type=int, default=32)
@corgie_option("--force_chunk_xy", nargs=1, type=int, default=None)
@corgie_option("--mip", nargs=1, type=int, default=0)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def vote(
    ctx,
    src_layer_spec,
    dst_folder,
    chunk_xy,
    mip,
    z_offsets,
    force_chunk_xy,
    consensus_threshold,
    blur_sigma,
    kernel_size,
    start_coord,
    end_coord,
    coord_mip,
    suffix,
):

    scheduler = ctx.obj["scheduler"]

    if suffix is None:
        suffix = "_voted"
    else:
        suffix = f"_{suffix}"

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)
    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    if force_chunk_xy is None:
        force_chunk_xy = chunk_xy

    dst_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=dst_folder,
        name="dst",
        types=["field", "float_tensor"],
        readonly=False,
        suffix=suffix,
        force_chunk_xy=force_chunk_xy,
        overwrite=True,
    )

    vote_weights = dst_stack.create_sublayer(
        name="vote_weights",
        layer_type="float_tensor",
        overwrite=True,
        num_channels=max(len(src_stack), len(z_offsets)),
    )
    voted_field = dst_stack.create_sublayer(
        name="voted_field", layer_type="field", overwrite=True
    )

    vote_stitch_job = VoteJob(
        input_fields=src_stack.get_layers(),
        output_field=voted_field,
        chunk_xy=chunk_xy,
        bcube=bcube,
        z_offsets=z_offsets,
        mip=mip,
        consensus_threshold=consensus_threshold,
        blur_sigma=blur_sigma,
        kernel_size=kernel_size,
        weights_layer=vote_weights,
    )
    scheduler.register_job(
        vote_stitch_job, job_name=f"Vote {bcube}",
    )

    scheduler.execute_until_completion()
    corgie_logger.debug("Done!")




import json
import copy
from math import floor, ceil
import numpy as np

from corgie.helpers import crop


def get_bcube_from_coords(start_coord, end_coord, coord_mip, cant_be_empty=True):
    xs, ys, zs = [int(i) for i in start_coord.split(",")]
    xe, ye, ze = [int(i) for i in end_coord.split(",")]
    bcube = BoundingCube(xs, xe, ys, ye, zs, ze, coord_mip)

    if cant_be_empty and bcube.area() * bcube.z_size() == 0:
        raise Exception(
            "Attempted creation of an empty bounding "
            "when 'cant_be_empty' flag is set to True"
        )

    return bcube


def get_bcube_from_vertices(vertices, resolution, mip, cant_be_empty=True):
    bcube_args = []
    for dim in range(3):
        dim_min = int(min(vertices[:, dim]) // resolution[dim])
        dim_max = int(max(vertices[:, dim]) // resolution[dim]) + 1
        bcube_args.append(dim_min)
        bcube_args.append(dim_max)

    bcube = BoundingCube(*bcube_args, mip)
    if cant_be_empty and bcube.area() * bcube.z_size() == 0:
        raise Exception(
            "Attempted creation of an empty bounding "
            "when 'cant_be_empty' flag is set to True"
        )

    return bcube


class BoundingCube:
    def __init__(self, xs, xe, ys, ye, zs, ze, mip):
        self.m0_x = (None, None)
        self.m0_y = (None, None)
        self.z = (None, None)
        self.reset_coords(xs, xe, ys, ye, zs, ze, mip=mip)

    # TODO
    # def contains(self, other):
    # def insets(self, other, mip):

    def get_bounding_pts(self):
        return (self.m0_x[0], self.m0_y[0], self.z[0]), (
            self.m0_x[1],
            self.m0_y[1],
            self.z[1],
        )

    def contains(self, other):
        if self.m0_y[1] < other.m0_y[1]:
            return False
        if self.m0_x[1] < other.m0_x[1]:
            return False
        if self.z[1] < other.z[1]:
            return False

        if other.m0_x[0] < self.m0_x[0]:
            return False
        if other.m0_y[0] < self.m0_y[0]:
            return False
        if other.z[0] < self.z[0]:
            return False

        return True

    # TODO: delete?
    def intersects(self, other):
        assert type(other) == type(self)
        if other.m0_x[1] < self.m0_x[0]:
            return False
        if other.m0_y[1] < self.m0_y[0]:
            return False
        if self.m0_x[1] < other.m0_x[0]:
            return False
        if self.m0_y[1] < other.m0_y[0]:
            return False
        if self.z[1] < other.z[0]:
            return False
        if other.z[1] < self.z[0]:
            return False
        return True

    def reset_coords(
        self, xs=None, xe=None, ys=None, ye=None, zs=None, ze=None, mip=0, in_place=True
    ):

        if in_place:
            target_cube = self
        else:
            target_cube = copy.deepcopy(self)
        scale_factor = 2 ** mip
        if xs is not None:
            target_cube.m0_x = (int(xs * scale_factor), target_cube.m0_x[1])
        if xe is not None:
            target_cube.m0_x = (target_cube.m0_x[0], int(xe * scale_factor))

        if ys is not None:
            target_cube.m0_y = (int(ys * scale_factor), target_cube.m0_y[1])
        if ye is not None:
            target_cube.m0_y = (target_cube.m0_y[0], int(ye * scale_factor))

        if zs is not None:
            target_cube.z = (int(zs), target_cube.z[1])
        if ze is not None:
            target_cube.z = (target_cube.z[0], int(ze))

        return target_cube

    def get_offset(self, mip=0):
        scale_factor = 2 ** mip
        return (
            self.m0_x[0] / scale_factor + self.x_size(mip=0) / 2 / scale_factor,
            self.m0_y[0] / scale_factor + self.y_size(mip=0) / 2 / scale_factor,
        )

    def x_range(self, mip):
        scale_factor = 2 ** mip
        xl = int(round((self.m0_x[1] - self.m0_x[0]) / scale_factor))
        xs = floor(self.m0_x[0] / scale_factor)
        return [xs, xs + xl]

    def y_range(self, mip):
        scale_factor = 2 ** mip
        yl = int(round((self.m0_y[1] - self.m0_y[0]) / scale_factor))
        ys = floor(self.m0_y[0] / scale_factor)
        return [ys, ys + yl]

    def z_range(self):
        return list(copy.deepcopy(self.z))

    def area(self, mip=0):
        x_size = self.x_size(mip)
        y_size = self.y_size(mip)
        return x_size * y_size

    def x_size(self, mip):
        x_range = self.x_range(mip)
        return int(x_range[1] - x_range[0])

    def y_size(self, mip):
        y_range = self.y_range(mip)
        return int(y_range[1] - y_range[0])

    def z_size(self):
        return int(self.z[1] - self.z[0])

    def minpt(self, mip=0):
        return np.array([self.x_range(mip)[0], self.y_range(mip)[0], self.z_range()[0]])

    def maxpt(self, mip=0):
        return np.array([self.x_range(mip)[1], self.y_range(mip)[1], self.z_range()[1]])

    def to_filename(self, mip=0):
        minpt = self.minpt(mip)
        maxpt = self.maxpt(mip)
        return "_".join((str(minpt[i]) + "-" + str(maxpt[i]) for i in range(3)))

    @property
    def size(self, mip=0):
        return self.x_size(mip=mip), self.y_size(mip=mip), self.z_size()

    def crop(self, crop_xy, mip):
        scale_factor = 2 ** mip
        m0_crop_xy = crop_xy * scale_factor
        self.set_m0(
            self.m0_x[0] + m0_crop_xy,
            self.m0_x[1] - m0_crop_xy,
            self.m0_y[0] + m0_crop_xy,
            self.m0_y[1] - m0_crop_xy,
        )

    def uncrop(self, crop_xy, mip):
        """Uncrop the bounding box by crop_xy at given MIP level"""
        scale_factor = 2 ** mip
        m0_crop_xy = crop_xy * scale_factor
        result = self.clone()
        result.reset_coords(
            xs=self.m0_x[0] - m0_crop_xy,
            xe=self.m0_x[1] + m0_crop_xy,
            ys=self.m0_y[0] - m0_crop_xy,
            ye=self.m0_y[1] + m0_crop_xy,
            mip=0,
        )
        return result

    def zeros(self, mip):
        return np.zeros(
            (self.x_size(mip), self.y_size(mip), self.z_size()), dtype=np.float32
        )

    def x_res_displacement(self, d_pixels, mip):
        disp_prop = d_pixels / self.x_size(mip=0)
        result = np.full(
            (self.x_size(mip), self.y_size(mip)), disp_prop, dtype=np.float32
        )
        return result

    def y_res_displacement(self, d_pixels, mip):
        disp_prop = d_pixels / self.y_size(mip=0)
        result = np.full(
            (self.x_size(mip), self.y_size(mip)), disp_prop, dtype=np.float32
        )
        return result

    def spoof_x_y_residual(self, x_d, y_d, mip, crop_amount=0):
        x_res = crop(self.x_res_displacement(x_d, mip=mip), crop_amount)
        y_res = crop(self.y_res_displacement(y_d, mip=mip), crop_amount)
        result = np.stack((x_res, y_res), axis=2)
        result = np.expand_dims(result, 0)
        return result

    def __eq__(self, x):
        if isinstance(x, BoundingCube):
            return (self.m0_x == x.m0_x) and (self.m0_y == x.m0_y) and (self.z == x.z)
        return False

    def __str__(self, mip=0):
        return "[MIP {}] {}, {}, {}".format(
            mip, self.x_range(mip), self.y_range(mip), self.z_range()
        )

    def __repr__(self):
        return self.__str__(mip=0)

    def translate_v1(self, x=0, y=0, z=0, mip=0):
        assert isinstance(x, int)
        assert isinstance(y, int)
        assert isinstance(z, int)

        scale_factor = 2 ** mip

        return BoundingCube(
            xs=self.m0_x[0] + x,
            xe=self.m0_x[1] + x,
            ys=self.m0_y[0] + y,
            ye=self.m0_y[1] + y,
            zs=self.z[0] + z,
            ze=self.z[1] + z,
            mip=mip,
        )

    def clone(self):
        return copy.deepcopy(self)

    def copy(self):
        return self.clone()

    def translate(self, z_offset=0, x_offset=0, y_offset=0, mip=0):
        x_range = self.x_range(mip=mip)
        y_range = self.y_range(mip=mip)
        z_range = self.z_range()
        return BoundingCube(
            xs=x_range[0] + x_offset,
            xe=x_range[1] + x_offset,
            ys=y_range[0] + y_offset,
            ye=y_range[1] + y_offset,
            zs=z_range[0] + z_offset,
            ze=z_range[1] + z_offset,
            mip=mip,
        )

    def to_slices(self, zs, ze=None, mip=0):
        x_range = self.x_range(mip=mip)
        y_range = self.y_range(mip=mip)
        return slice(*x_range), slice(*y_range), slice(*self.z)



import click
from cloudfiles import CloudFiles
from cloudvolume import PrecomputedSkeleton
from collections import defaultdict
from copy import deepcopy
import kimimaro
import numpy as np
import pickle

from corgie import scheduling, argparsers, helpers, stack
from corgie.log import logger as corgie_logger
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)


class SkeletonJob(scheduling.Job):
    def __init__(
        self,
        seg_layer,
        dst_path,
        timestamp,
        bcube,
        chunk_xy,
        chunk_z,
        mip,
        teasar_params,
        object_ids=None,
        fix_branching=True,
        fix_borders=True,
        fix_avocados=False,
        dust_threshold=0,
        tick_threshold=1000,
        single_merge_mode=True,
    ):
        self.seg_layer = seg_layer
        self.dst_path = dst_path
        self.timestamp = timestamp
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.mip = mip
        self.teasar_params = teasar_params
        self.object_ids = object_ids
        self.fix_branching = fix_branching
        self.fix_borders = fix_borders
        self.fix_avocados = fix_avocados
        self.dust_threshold = dust_threshold
        self.tick_threshold = tick_threshold
        self.single_merge_mode = single_merge_mode
        super().__init__()

    def task_generator(self):
        chunks = self.seg_layer.break_bcube_into_chunks(
            bcube=self.bcube,
            chunk_xy=self.chunk_xy,
            chunk_z=self.chunk_z,
            mip=self.mip,
            readonly=True,
        )
        tasks = [
            SkeletonTask(
                self.seg_layer,
                self.dst_path,
                timestamp=self.timestamp,
                mip=self.mip,
                teasar_params=self.teasar_params,
                object_ids=self.object_ids,
                dust_threshold=self.dust_threshold,
                fix_branching=self.fix_branching,
                fix_borders=self.fix_borders,
                fix_avocados=self.fix_avocados,
                bcube=input_chunk,
            )
            for input_chunk in chunks
        ]
        corgie_logger.info(
            f"Yielding skeletonization tasks for bcube: {self.bcube}, MIP: {self.mip}"
        )
        yield tasks
        yield scheduling.wait_until_done
        if self.single_merge_mode:
            merge_tasks = [
                MergeSkeletonTask(
                    self.dst_path,
                    self.mip,
                    self.dust_threshold,
                    self.tick_threshold,
                    str(object_id),
                )
                for object_id in self.object_ids
            ]
            yield merge_tasks
        else:
            yield [
                MergeSkeletonTask(
                    self.dst_path, self.mip, self.dust_threshold, self.tick_threshold
                )
            ]


class SkeletonTask(scheduling.Task):
    def __init__(
        self,
        seg_layer,
        dst_path,
        timestamp,
        bcube,
        mip,
        teasar_params,
        object_ids,
        fix_branching,
        fix_borders,
        fix_avocados,
        dust_threshold,
    ):
        super().__init__(self)
        self.seg_layer = seg_layer
        self.dst_path = dst_path
        self.timestamp = timestamp
        self.bcube = bcube
        self.mip = mip
        self.teasar_params = teasar_params
        self.object_ids = object_ids
        self.fix_branching = fix_branching
        self.fix_borders = fix_borders
        self.fix_avocados = fix_avocados
        self.dust_threshold = dust_threshold

    def execute(self):
        corgie_logger.info(
            f"Skeletonizing {self.seg_layer} at MIP{self.mip}, region: {self.bcube}"
        )
        seg_data = self.seg_layer.read(
            bcube=self.bcube, mip=self.mip, timestamp=self.timestamp
        )
        resolution = self.seg_layer.cv[self.mip].resolution
        skeletons = kimimaro.skeletonize(
            seg_data,
            self.teasar_params,
            object_ids=self.object_ids,
            anisotropy=resolution,
            dust_threshold=self.dust_threshold,
            progress=False,
            fix_branching=self.fix_branching,
            fix_borders=self.fix_borders,
            fix_avocados=self.fix_avocados,
        ).values()

        minpt = self.bcube.minpt(self.mip)
        for skel in skeletons:
            skel.vertices[:] += minpt * resolution

        cf = CloudFiles(self.dst_path)
        for skel in skeletons:
            path = "{}:{}".format(skel.id, self.bcube.to_filename(self.mip))
            cf.put(
                path=path,
                content=pickle.dumps(skel),
                compress="gzip",
                content_type="application/python-pickle",
                cache_control=False,
            )


class MergeSkeletonTask(scheduling.Task):
    def __init__(self, dst_path, mip, dust_threshold, tick_threshold, prefix=""):
        super().__init__(self)
        self.dst_path = dst_path
        self.cf = CloudFiles(self.dst_path)
        self.mip = mip
        self.dust_threshold = dust_threshold
        self.tick_threshold = tick_threshold
        self.prefix = prefix

    def execute(self):
        corgie_logger.info(f"Merging skeletons at {self.dst_path}")
        fragment_filenames = self.cf.list(prefix=self.prefix, flat=True)
        skeleton_files = self.cf.get(fragment_filenames)
        skeletons = defaultdict(list)
        for skeleton_file in skeleton_files:
            try:
                colon_index = skeleton_file["path"].index(":")
            except ValueError:
                # File is full skeleton, not fragment
                continue
            seg_id = skeleton_file["path"][0:colon_index]
            skeleton_fragment = pickle.loads(skeleton_file["content"])
            if not skeleton_fragment.empty():
                skeletons[seg_id].append(skeleton_fragment)
        for seg_id, skeleton_fragments in skeletons.items():
            skeleton = PrecomputedSkeleton.simple_merge(
                skeleton_fragments
            ).consolidate()
            skeleton = kimimaro.postprocess(
                skeleton, self.dust_threshold, self.tick_threshold
            )
            skeleton.id = int(seg_id)
            self.cf.put(path=seg_id, content=skeleton.to_precomputed(), compress="gzip")
            corgie_logger.info(f"Finished skeleton {seg_id}")


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--seg_layer_spec",
    "--s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Seg layer from which to skeletonize segments.",
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where to store the skeletons",
)
@corgie_option(
    "--timestamp",
    nargs=1,
    type=int,
    default=None,
    help="UNIX timestamp that specifies which segmentation to use. Only relevant for graphene segmentations",
)
@corgie_optgroup("Skeletonization Parameters")
@corgie_option("--mip", nargs=1, type=int, required=True)
@corgie_option("--teasar_scale", nargs=1, type=int, default=10)
@corgie_option("--teasar_const", nargs=1, type=int, default=10)
@corgie_option("--ids", multiple=True, type=int, help="Segmentation ids to skeletonize")
@corgie_option(
    "--ids_filepath", type=str, help="File containing segmentation ids to skeletonize"
)
@corgie_option("--tick_threshold", nargs=1, type=int, default=1000)
@corgie_option("--chunk_xy", nargs=1, type=int, default=256)
@corgie_option("--chunk_z", nargs=1, type=int, default=256)
@corgie_option(
    "--single_merge_mode",
    nargs=1,
    type=bool,
    default=True,
    help="Set to True to have 1 skeleton merge=1 task",
)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def create_skeletons(
    ctx,
    seg_layer_spec,
    dst_folder,
    timestamp,
    mip,
    teasar_scale,
    teasar_const,
    ids,
    ids_filepath,
    tick_threshold,
    chunk_xy,
    chunk_z,
    single_merge_mode,
    start_coord,
    end_coord,
    coord_mip,
):
    scheduler = ctx.obj["scheduler"]

    corgie_logger.debug("Setting up layers...")
    seg_stack = create_stack_from_spec(seg_layer_spec, name="src", readonly=True)
    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    object_ids = ids
    if ids_filepath is not None:
        object_ids = []
        with open(ids_filepath, "r") as f:
            line = f.readline()
            while line:
                object_ids.append(int(line))
                line = f.readline()
    if object_ids is None or len(object_ids) == 0:
        raise ValueError("Must specify ids to skeletonize")
    object_ids = list(object_ids)
    teasar_params = {"scale": teasar_scale, "const": teasar_const}

    seg_layer = seg_stack.get_layers_of_type("segmentation")[0]
    skeleton_job = SkeletonJob(
        seg_layer=seg_layer,
        dst_path=dst_folder,
        timestamp=timestamp,
        bcube=bcube,
        chunk_xy=chunk_xy,
        chunk_z=chunk_z,
        mip=mip,
        teasar_params=teasar_params,
        object_ids=object_ids,
        tick_threshold=tick_threshold,
        single_merge_mode=single_merge_mode,
    )

    scheduler.register_job(skeleton_job, job_name="Skeletonize {}".format(bcube))

    scheduler.execute_until_completion()
    result_report = f"Skeletonized {str(seg_layer)}. "
    corgie_logger.info(result_report)



from dataclasses import astuple, dataclass

import numpy as np
import torch
from copy import deepcopy


class Binarizer:
    def __init__(self, binarization):
        self.bin = binarization

    def __call__(self, tens):
        if self.bin is None:
            return tens
        elif self.bin[0] == "neq":
            return tens != self.bin[1]
        elif self.bin[0] == "eq":
            return tens == self.bin[1]
        elif self.bin[0] == "gt":
            return tens > self.bin[1]
        elif self.bin[0] == "lt":
            return tens < self.bin[1]


class PartialSpecification:
    def __init__(self, f, **kwargs):
        self.f = f
        self.constr_kwargs = kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __call__(self, **kwargs):
        return self.f(**self.constr_kwargs, **kwargs)

    def __getitem__(self, k):
        return self.constr_kwargs[k]


@dataclass
class Translation:
    x: float
    y: float

    def __iter__(self):
        return iter(astuple(self))

    def __add__(self, T):
        return Translation(self.x + T.x, self.y + T.y)

    def __sub__(self, T):
        return Translation(self.x - T.x, self.y - T.y)

    def __mul__(self, scalar):
        return Translation(self.x * scalar, self.y * scalar)

    def __rmul__(self, scalar):
        return self.__mul__(scalar)

    def __floordiv__(self, scalar):
        return Translation(self.x // scalar, self.y // scalar)

    def __truediv__(self, scalar):
        return Translation(self.x / scalar, self.y / scalar)

    def to_tensor(self, **kwargs):
        return torch.tensor([[[[self.x]], [[self.y]]]], **kwargs)

    def round(self, ndigits=None):
        return Translation(round(self.x, ndigits), round(self.y, ndigits))

    def copy(self):
        return deepcopy(self)

    def round_to_mip(self, src_mip, tgt_mip):
        if tgt_mip is None:
            return self.copy()
        elif tgt_mip <= src_mip:
            return Translation(self.x, self.y)  # Return copy
        else:
            snap_factor = 2 ** (tgt_mip - src_mip)
            return (self // snap_factor) * snap_factor


def percentile_trans_adjuster(field, h=25, l=75, unaligned_img=None):
    if field is None:
        result = Translation(0, 0)
    else:
        nonzero_field_mask = (field[:, 0] != 0) & (field[:, 1] != 0)

        if unaligned_img is not None:
            no_tissue = field.field().from_pixels()(unaligned_img) == 0
            nonzero_field_mask[..., no_tissue.squeeze()] = False

        nonzero_field = field[..., nonzero_field_mask.squeeze()].squeeze()

        if nonzero_field.sum() == 0:
            result = Translation(0, 0)
        else:
            low_l = percentile(nonzero_field, l)
            high_l = percentile(nonzero_field, h)
            mid = 0.5 * (low_l + high_l)
            result = Translation(x=int(mid[0]), y=int(mid[1]))

    return result


def percentile(field, q):
    # https://gist.github.com/spezold/42a451682422beb42bc43ad0c0967a30
    """
    Return the ``q``-th percentile of the flattened input tensor's data.
    CAUTION:
     * Needs PyTorch >= 1.1.0, as ``torch.kthvalue()`` is used.
     * Values are not interpolated, which corresponds to
       ``numpy.percentile(..., interpolation="nearest")``.
    :param field: Input tensor.
    :param q: Percentile to compute, which must be between 0 and 100 inclusive.
    :return: Resulting value (scalar).
    """
    # Note that ``kthvalue()`` works one-based, i.e. the first sorted value
    # indeed corresponds to k=1, not k=0! Use float(q) instead of q directly,
    # so that ``round()`` returns an integer, even if q is a np.float32.
    k = 1 + round(0.01 * float(q) * (field.shape[1] - 1))
    result = field.kthvalue(k, dim=1).values
    return result


def crop(**kwargs):
    raise NotImplementedError


def expand_to_dims(tens, dims):
    tens_dims = len(tens.shape)
    assert (tens_dims) <= dims
    tens = tens[(None,) * (dims - tens_dims)]
    return tens


def cast_tensor_type(tens, dtype):
    """
    tens: pytorch tens
    dtype: string, eg 'float', 'int', 'byte'
    """
    if dtype is not None:
        assert hasattr(tens, dtype)
        return getattr(tens, dtype)()
    else:
        return tens


def read_mask_list(mask_list, bcube, mip):
    result = None

    for m in mask_list:
        this_data = m.read(bcube=bcube, mip=mip).to(torch.bool)
        if result is None:
            result = this_data
        else:
            result = result | this_data

    return result


def crop(data, c):
    if c == 0:
        return data
    else:
        if data.shape[-1] == data.shape[-2]:
            return data[..., c:-c, c:-c]
        elif data.shape[-2] == data.shape[-3] and data.shape[-1] == 2:  # field
            return data[..., c:-c, c:-c, :]


def coarsen_mask(mask, n=1, flip=False):
    kernel = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
    kernel_var = (
        torch.FloatTensor(kernel).unsqueeze(0).unsqueeze(0).to(mask.device).float()
    )
    k = torch.nn.Parameter(data=kernel_var, requires_grad=False)
    for _ in range(n):
        if flip:
            mask = mask.logical_not()
        mask = torch.nn.functional.conv2d(mask.float(), kernel_var, padding=1) > 1
        if flip:
            mask = mask.logical_not()
        mask = mask

    return mask


def zeros(*args, **kwargs):
    return torch.zeros(*args, **kwargs)


class BoolFn:
    def __init__(self, exp):
        """Create a boolean expression from dict describing a linear threshold (LT) neuron

        A LT neuron is described as dict with keys "inputs" and "threshold". The LT neuron
        will evaluate to true if the sum of "inputs" is greater than or equal to the
        "threshold". The "inputs" of an LT neuron maybe include other LT neurons. Final
        variables must be expressed as a dict that does not contain the key "inputs". These
        variables are stored as BoolVars. To evaluate the Boolean expression, pass a method
        that can evaluate the dict for each BoolVar.

        Args:
            exp (dict of dicts): dict describing nested LT neurons. For example:
            ```
            {
            "inputs": [
                {
                "inputs": [
                    { "weight": 1, "key": "a", "offset": 0 },
                    { "weight": -1, "key": "b", "offset": -1 },
                    { "weight": 1, "key": "c", "offset": 2 }
                ],
                "threshold": 2
                },
                {
                "inputs": [
                    { "weight": 1, "key": "a", "offset": 1 },
                    { "weight": 1, "key": "c", "offset": 0 }
                ],
                "threshold": 1
                }
            ],
            "threshold": 0
            }
            ```

            evaluates as
            `((a[0] - b[-1] + c[2] > 2) + (a[1] - c[0]) > 1) > 0`.
        """
        self.inputs = [
            BoolFn(e) if "inputs" in e else BoolVar(e) for e in exp["inputs"]
        ]
        self.threshold = exp["threshold"]

    def __call__(self, fn):
        return sum([i(fn) for i in self.inputs]) > self.threshold


class BoolVar:
    def __init__(self, exp):
        """See description of BoolFn"""
        self.exp = exp

    def __call__(self, fn):
        return fn(self.exp)


import click
import procspec

from corgie.log import logger as corgie_logger
from corgie.boundingcube import get_bcube_from_coords
from corgie import scheduling, helpers, stack
from corgie.argparsers import (
    LAYER_HELP_STR,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)

class MultiSectionCompareJob(scheduling.Job):
    """Names of img layers in dst_stack should be z_offsets
    """
    def __init__(
        self,
        src_stack,
        dst_stack,
        chunk_xy,
        processor_spec,
        mip,
        dst_mip,
        pad,
        bcube,
        tgt_stack=None,
        suffix="",
    ):

        self.src_stack = src_stack
        if tgt_stack is None:
            tgt_stack = src_stack

        self.tgt_stack = tgt_stack
        self.dst_stack = dst_stack
        self.chunk_xy = chunk_xy
        self.pad = pad
        self.bcube = bcube

        self.suffix = suffix

        self.processor_spec = processor_spec
        self.mip = mip
        self.dst_mip = dst_mip
        super().__init__()

    def task_generator(self):
        chunks = self.dst_stack.get_layers()[0].break_bcube_into_chunks(
            bcube=self.bcube,
            chunk_xy=self.chunk_xy,
            chunk_z=1,
            mip=self.dst_mip,
            return_generator=True,
        )

        tasks = (
            MultiSectionCompareTask(
                processor_spec=self.processor_spec,
                src_stack=self.src_stack,
                tgt_stack=self.tgt_stack,
                dst_stack=self.dst_stack,
                mip=self.mip,
                dst_mip=self.dst_mip,
                pad=self.pad,
                bcube=input_chunk,
            )
            for input_chunk in chunks
        )
        print(
            f"Yielding MultiSectionCompareTasks for bcube: {self.bcube}, MIP: {self.mip}"
        )

        yield tasks

class MultiSectionCompareTask(scheduling.Task):
    def __init__(
        self,
        processor_spec,
        src_stack,
        tgt_stack,
        dst_stack,
        mip,
        dst_mip,
        pad,
        bcube,
    ):
        super().__init__()
        self.processor_spec = processor_spec
        self.src_stack = src_stack
        self.tgt_stack = tgt_stack
        self.dst_stack = dst_stack
        self.mip = mip
        self.dst_mip = dst_mip
        self.pad = pad
        self.bcube = bcube

    def execute(self):
        tgt_bcube = self.bcube.uncrop(self.pad, self.mip)
        _, tgt_data_dict = self.tgt_stack.read_data_dict(
            tgt_bcube, mip=self.mip, stack_name="tgt"
        )

        dst_layers = self.dst_stack.get_layers_of_type("img")
        for dst_layer in dst_layers:
            z_offset = int(dst_layer.name)
            src_bcube = tgt_bcube.translate(z_offset=z_offset)
            processor = procspec.parse_proc(spec_str=self.processor_spec)

            _, src_data_dict = self.src_stack.read_data_dict(
                src_bcube, mip=self.mip, stack_name="src"
            )

            processor_input = {**src_data_dict, **tgt_data_dict}

            result = processor(processor_input, output_key="result")
            crop = self.pad // 2 ** (self.dst_mip - self.mip)
            cropped_result = helpers.crop(result, crop)
            dst_layer.write(cropped_result, bcube=self.bcube, mip=self.dst_mip)


@click.command()
# Layers
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where aligned stack will go",
)
@corgie_optgroup("Multi Section Compare Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--pad", nargs=1, type=int, default=256)
@corgie_option("--crop", nargs=1, type=int, default=None)
@corgie_option("--force_chunk_xy", nargs=1, type=int, default=None)
@corgie_option("--z_offsets", multiple=True, type=int, default=[-1])
@corgie_option("--processor_spec", nargs=1, type=str, required=True, multiple=False)
@corgie_option("--processor_mip", "-m", nargs=1, type=int, required=True, multiple=True)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def multi_section_compare(
    ctx,
    src_layer_spec,
    dst_folder,
    chunk_xy,
    pad,
    crop,
    force_chunk_xy,
    z_offsets,
    processor_spec,
    processor_mip,
    start_coord,
    end_coord,
    coord_mip,
):

    scheduler = ctx.obj["scheduler"]

    if crop is None:
        crop = pad

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)
    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    if force_chunk_xy is None:
        force_chunk_xy = chunk_xy

    dst_stack = stack.Stack(name="dst", folder=dst_folder) 
    dst_stack.reference_layer = src_stack.get_layers()[0]

    for z_offset in z_offsets:
        dst_stack.create_sublayer(
            name=z_offset,
            layer_type="img",
            dtype="float32",
            force_chunk_xy=force_chunk_xy,
            overwrite=True,
        )

    multi_section_compare_job = MultiSectionCompareJob(
        src_stack=src_stack,
        dst_stack=dst_stack,
        chunk_xy=chunk_xy,
        bcube=bcube,
        pad=pad,
        processor_spec=processor_spec,
        mip=processor_mip[0],
        dst_mip=processor_mip[0],
    )
    scheduler.register_job(
        multi_section_compare_job, job_name=f"MultiSectionCompare {bcube}",
    )

    scheduler.execute_until_completion()
    corgie_logger.debug("Done!")




import click
from copy import deepcopy

from corgie import scheduling, argparsers, helpers, stack

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords

from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)

from corgie.cli.render import RenderJob
from corgie.cli.copy import CopyJob
from corgie.cli.vote import VoteJob
from corgie.cli.compute_field import ComputeFieldJob
from corgie.cli.seethrough import SeethroughCompareJob

from corgie.cli.downsample import DownsampleJob
from corgie.cli.upsample import UpsampleJob


class AlignBlockJob(scheduling.Job):
    def __init__(
        self,
        src_stack,
        dst_stack,
        cf_method,
        render_method,
        bcube,
        seethrough_method=None,
        copy_start=True,
        backward=False,
        vote_dist=1,
        suffix=None,
        consensus_threshold=3.,
        blur_sigma=15,
        kernel_size=32,
        use_starters=True,
    ):
        """Align block with and without voting

        Args:
            use_starters (bool): whether to use starter sections or not
        """
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.bcube = bcube
        self.seethrough_method = seethrough_method

        self.cf_method = cf_method
        self.render_method = render_method

        self.copy_start = copy_start
        self.use_starters = use_starters
        self.backward = backward
        self.vote_dist = vote_dist
        self.suffix = suffix

        self.consensus_threshold = consensus_threshold
        self.blur_sigma = blur_sigma
        self.kernel_size = kernel_size

        self.tgt_stack = deepcopy(self.dst_stack)
        field_dir = f"field{self.suffix}"
        self.final_field = self.dst_stack.create_sublayer(
            field_dir, layer_type="field", overwrite=True
        )
        self.estimated_fields = {}
        self.z_offsets = range(1, self.vote_dist + 1)

        if self.backward:
            z_step = -1
        else:
            z_step = 1

        for k in self.z_offsets:
            offset = -k * z_step
            f = self.dst_stack.create_sublayer(
                f"{field_dir}/{offset}", layer_type="field", overwrite=True
            )
            self.estimated_fields[offset] = f

        if self.seethrough_method is not None:
            self.seethrough_mask_layer = self.dst_stack.create_sublayer(
                f"seethrough_mask{self.suffix}",
                layer_type="mask",
                overwrite=True,
            )
            self.pixel_offset_layer = self.tgt_stack.create_unattached_sublayer(
                f"pixel_offset{self.suffix}", layer_type="img", overwrite=True
            )
        else:
            self.seethrough_mask_layer = None
            self.pixel_offset_layer = None

        super().__init__()

    def task_generator(self):
        if not self.backward:
            z_step = 1
            z_start = self.bcube.z_range()[0]
            z_end = self.bcube.z_range()[1]
        else:
            z_step = -1
            z_start = self.bcube.z_range()[1]
            z_end = self.bcube.z_range()[0]
        seethrough_offset = -z_step
        if (self.vote_dist > 1) and self.use_starters:
            starter_section_start = z_start - (self.vote_dist - 1) * z_step
        else:
            starter_section_start = z_start
        z_range = range(starter_section_start, z_end, z_step)

        processor_mips = set(self.cf_method.processor_mip)
        processor_and_seethrough_mips = processor_mips.copy()
        if self.seethrough_method is not None:
            processor_and_seethrough_mips.add(self.seethrough_method.mip)

        corgie_logger.debug(
            f"Serial alignment, {z_start}->{z_end}, use_starters={self.use_starters}"
        )
        for z in z_range:
            bcube = self.bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
            # COPY FIRST SECTION OF THE BLOCK
            if (z == z_start) and self.copy_start:
                corgie_logger.debug(f"Copy section {z}")
                render_job = self.render_method(
                    src_stack=self.src_stack,
                    dst_stack=self.tgt_stack,
                    bcube=bcube,
                    mips=processor_and_seethrough_mips,
                    blackout_masks=True,
                )

                yield from render_job.task_generator
                yield scheduling.wait_until_done
            # CREATE STARTER SECTIONS FOR VOTING
            elif (z_step > 0 and z < z_start) or (z_step < 0 and z > z_start):
                corgie_logger.debug(f"Compute field vote starter {z_start}<{z}")
                offset = z_start - z
                compute_field_job = self.cf_method(
                    src_stack=self.src_stack,
                    tgt_stack=self.src_stack,
                    bcube=bcube,
                    tgt_z_offset=offset,
                    suffix=self.suffix,
                    dst_layer=self.final_field,
                )
                yield from compute_field_job.task_generator
                yield scheduling.wait_until_done

                corgie_logger.debug(f"Render vote starter {z_start}<{z}")
                render_job = self.render_method(
                    src_stack=self.src_stack,
                    dst_stack=self.tgt_stack,
                    bcube=bcube,
                    blackout_masks=True,
                    seethrough_mask_layer=None,  # No seethrough for starter sections
                    mips=self.cf_method.processor_mip,
                    additional_fields=[self.final_field],
                )
                yield from render_job.task_generator
                yield scheduling.wait_until_done

            # SERIAL ALIGNMENT (w/ or w/o voting)
            else:
                if self.vote_dist > 1:
                    for k in self.z_offsets:
                        offset = -k * z_step
                        corgie_logger.debug(f"Compute field {z+offset}<{z}")
                        compute_field_job = self.cf_method(
                            src_stack=self.src_stack,
                            tgt_stack=self.tgt_stack,
                            bcube=bcube,
                            tgt_z_offset=offset,
                            suffix=self.suffix,
                            dst_layer=self.estimated_fields[offset],
                        )

                        yield from compute_field_job.task_generator

                    yield scheduling.wait_until_done
                elif self.vote_dist == 1:
                    offset = -z_step
                    corgie_logger.debug(f"Compute final field field {z+offset}<{z}")
                    compute_field_job = self.cf_method(
                        src_stack=self.src_stack,
                        tgt_stack=self.tgt_stack,
                        bcube=bcube,
                        tgt_z_offset=offset,
                        suffix=self.suffix,
                        dst_layer=self.final_field,
                    )

                    yield from compute_field_job.task_generator
                    yield scheduling.wait_until_done
                else:
                    raise Exception()

                if self.vote_dist > 1:
                    corgie_logger.debug(f"Vote {z}")
                    chunk_xy = self.cf_method["chunk_xy"]
                    mip = self.cf_method["processor_mip"][-1]
                    ordered_fields = [self.estimated_fields[-k*z_step] for k in self.z_offsets]
                    vote_job = VoteJob(
                        input_fields=ordered_fields,
                        output_field=self.final_field,
                        chunk_xy=chunk_xy,
                        bcube=bcube,
                        z_offsets=[0],
                        mip=mip,
                        consensus_threshold=self.consensus_threshold,
                        blur_sigma=self.blur_sigma,
                        kernel_size=self.kernel_size,
                    )

                    yield from vote_job.task_generator
                    yield scheduling.wait_until_done

                if self.seethrough_method is not None:
                    # This sequence can be bundled into a "seethrough render" job
                    # First, adjust the field to the appropriate MIP for seethrough rendering
                    if self.cf_method["processor_mip"][-1] < self.seethrough_method.mip:
                        downsample_job = DownsampleJob(
                            src_layer=self.final_field,
                            chunk_xy=self.cf_method["chunk_xy"],
                            chunk_z=1,
                            mip_start=self.cf_method["processor_mip"][-1],
                            mip_end=self.seethrough_method.mip,
                            bcube=bcube,
                        )
                        yield from downsample_job.task_generator
                        yield scheduling.wait_until_done
                    elif (
                        self.cf_method["processor_mip"][-1] > self.seethrough_method.mip
                    ):
                        upsample_job = UpsampleJob(
                            src_layer=self.final_field,
                            chunk_xy=self.cf_method["chunk_xy"],
                            chunk_z=1,
                            mip_start=self.cf_method["processor_mip"][-1],
                            mip_end=self.seethrough_method.mip,
                            bcube=bcube,
                        )
                        yield from upsample_job.task_generator
                        yield scheduling.wait_until_done

                    # Next, render the images at the md mip level
                    # This could mean a reduntant render step, but that's fine
                    render_job = self.render_method(
                        src_stack=self.src_stack,
                        dst_stack=self.tgt_stack,
                        bcube=bcube,
                        blackout_masks=True,
                        preserve_zeros=True,
                        additional_fields=[self.final_field],
                        mips=self.seethrough_method.mip,
                    )

                    yield from render_job.task_generator
                    yield scheduling.wait_until_done

                    # Now, we'll apply misalignment detection to produce a mask
                    # this mask will be used in the final render step
                    seethrough_mask_job = self.seethrough_method(
                        src_stack=self.tgt_stack,  # we're looking for misalignments in the final stack
                        bcube=bcube,
                        tgt_z_offset=-z_step,
                        suffix=self.suffix,
                        dst_layer=self.seethrough_mask_layer,
                        pixel_offset_layer=self.pixel_offset_layer,
                    )

                    yield from seethrough_mask_job.task_generator
                    yield scheduling.wait_until_done

                    # We'll downsample the mask to be available at all mip levels
                    downsample_job = DownsampleJob(
                        src_layer=self.seethrough_mask_layer,
                        chunk_xy=self.seethrough_method.chunk_xy,
                        chunk_z=1,
                        mip_start=self.seethrough_method.mip,
                        mip_end=max(self.cf_method.processor_mip),
                        bcube=bcube,
                    )
                    upsample_job = UpsampleJob(
                        src_layer=self.seethrough_mask_layer,
                        chunk_xy=self.seethrough_method.chunk_xy,
                        chunk_z=1,
                        mip_start=self.seethrough_method.mip,
                        mip_end=min(self.cf_method.processor_mip),
                        bcube=bcube,
                    )

                    if (
                        min(self.cf_method.processor_mip) < self.seethrough_method.mip
                        or max(self.cf_method.processor_mip)
                        > self.seethrough_method.mip
                    ):
                        yield from downsample_job.task_generator
                        yield from upsample_job.task_generator
                        yield scheduling.wait_until_done

                corgie_logger.debug(f"Render {z}")
                render_job = self.render_method(
                    src_stack=self.src_stack,
                    dst_stack=self.tgt_stack,
                    bcube=bcube,
                    blackout_masks=True,
                    seethrough_mask_layer=self.seethrough_mask_layer,
                    seethrough_offset=seethrough_offset,
                    mips=min(self.cf_method.processor_mip),
                    additional_fields=[self.final_field],
                )
                yield from render_job.task_generator
                yield scheduling.wait_until_done

                # We'll downsample the tgt_stack for one-pass
                if min(self.cf_method.processor_mip) != max(
                    self.cf_method.processor_mip
                ):
                    dst_layer = self.tgt_stack.get_layers_of_type("img")[0]
                    downsample_job = DownsampleJob(
                        src_layer=dst_layer,
                        chunk_xy=self.cf_method.chunk_xy,
                        chunk_z=1,
                        mip_start=min(self.cf_method.processor_mip),
                        mip_end=max(self.cf_method.processor_mip),
                        bcube=bcube,
                    )
                    yield from downsample_job.task_generator
                    yield scheduling.wait_until_done


@click.command()
# Layers
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where aligned stack will go",
)
@corgie_option("--suffix", nargs=1, type=str, default=None)
@corgie_optgroup("Render Method Specification")
@corgie_option(
    "--seethrough_spec",
    nargs=1,
    type=str,
    default=None,
    multiple=True,
    help="Seethrough method spec. Use multiple times to specify different methods (e.g. seethrough misalignments, seethrough black, etc.)",
)
@corgie_option(
    "--seethrough_limit",
    nargs=1,
    type=int,
    default=None,
    multiple=True,
    help="For each seethrough method, how many sections are allowed to be seenthrough. 0 or None means no limit.",
)
@corgie_option("--seethrough_spec_mip", nargs=1, type=int, default=None)
@corgie_option("--render_pad", nargs=1, type=int, default=512)
@corgie_option("--render_chunk_xy", nargs=1, type=int, default=1024)
@corgie_optgroup("Compute Field Method Specification")
@corgie_option("--processor_spec", nargs=1, type=str, required=True, multiple=True)
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--blend_xy", nargs=1, type=int, default=0)
@corgie_option("--force_chunk_xy", nargs=1, type=int, default=None)
@corgie_option("--pad", nargs=1, type=int, default=256)
@corgie_option("--crop", nargs=1, type=int, default=None)
@corgie_option("--processor_mip", "-m", nargs=1, type=int, required=True, multiple=True)
@corgie_option("--copy_start/--no_copy_start", default=True)
@corgie_option("--use_starters/--no_starters", default=True)
@corgie_option(
    "--mode",
    type=click.Choice(["forward", "backward", "bidirectional"]),
    default="forward",
)
@corgie_option("--vote_dist", nargs=1, type=int, default=1)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def align_block(
    ctx,
    src_layer_spec,
    dst_folder,
    vote_dist,
    render_pad,
    render_chunk_xy,
    processor_spec,
    pad,
    crop,
    processor_mip,
    chunk_xy,
    start_coord,
    end_coord,
    coord_mip,
    blend_xy,
    force_chunk_xy,
    suffix,
    copy_start,
    use_starters,
    seethrough_spec,
    seethrough_limit,
    seethrough_spec_mip,
    mode,
    chunk_z=1,
):
    scheduler = ctx.obj["scheduler"]

    if suffix is None:
        suffix = "_aligned"
    else:
        suffix = f"_{suffix}"

    if crop is None:
        crop = pad
    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)
    src_stack.folder = dst_folder

    if force_chunk_xy is None:
        force_chunk_xy = chunk_xy

    dst_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=dst_folder,
        name="dst",
        types=["img", "mask"],
        readonly=False,
        suffix=suffix,
        force_chunk_xy=force_chunk_xy,
        overwrite=True,
    )

    render_method = helpers.PartialSpecification(
        f=RenderJob,
        pad=render_pad,
        chunk_xy=render_chunk_xy,
        chunk_z=1,
        render_masks=False,
    )

    if seethrough_spec != tuple():
        assert seethrough_spec_mip is not None

        seethrough_method = helpers.PartialSpecification(
            f=SeethroughCompareJob,
            mip=seethrough_spec_mip,
            processor_spec=seethrough_spec,
            chunk_xy=chunk_xy,
            pad=pad,
            crop=pad,
            seethrough_limit=seethrough_limit,
        )
    else:
        seethrough_method = None

    cf_method = helpers.PartialSpecification(
        f=ComputeFieldJob,
        pad=pad,
        crop=crop,
        processor_mip=processor_mip,
        processor_spec=processor_spec,
        chunk_xy=chunk_xy,
        blend_xy=blend_xy,
        chunk_z=1,
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    if mode == "bidirectional":
        z_mid = (bcube.z_range()[1] + bcube.z_range()[0]) // 2
        bcube_back = bcube.reset_coords(ze=z_mid, in_place=False)
        bcube_forv = bcube.reset_coords(zs=z_mid, in_place=False)

        align_block_job_back = AlignBlockJob(
            src_stack=src_stack,
            dst_stack=dst_stack,
            bcube=bcube_back,
            render_method=render_method,
            cf_method=cf_method,
            seethrough_method=seethrough_method,
            suffix=suffix,
            copy_start=copy_start,
            backward=True,
            vote_dist=vote_dist,
            use_starters=use_starters,
        )
        scheduler.register_job(
            align_block_job_back,
            job_name="Backward Align Block {}".format(bcube),
        )

        align_block_job_forv = AlignBlockJob(
            src_stack=src_stack,
            dst_stack=deepcopy(dst_stack),
            bcube=bcube_forv,
            render_method=render_method,
            cf_method=cf_method,
            seethrough_method=seethrough_method,
            suffix=suffix,
            copy_start=True,
            backward=False,
            vote_dist=vote_dist,
            use_starters=use_starters,
        )
        scheduler.register_job(
            align_block_job_forv,
            job_name="Forward Align Block {}".format(bcube),
        )
    else:
        align_block_job = AlignBlockJob(
            src_stack=src_stack,
            dst_stack=dst_stack,
            bcube=bcube,
            render_method=render_method,
            cf_method=cf_method,
            seethrough_method=seethrough_method,
            suffix=suffix,
            copy_start=copy_start,
            backward=mode == "backward",
            vote_dist=vote_dist,
            use_starters=use_starters,
        )

        # create scheduler and execute the job
        scheduler.register_job(align_block_job, job_name="Align Block {}".format(bcube))

    scheduler.execute_until_completion()
    result_report = (
        f"Aligned layers {[str(l) for l in src_stack.get_layers_of_type('img')]}. "
        f"Results in {[str(l) for l in dst_stack.get_layers_of_type('img')]}"
    )
    corgie_logger.info(result_report)



from corgie.layers.base import get_layer_types, str_to_layer_type

# when adding a new layer type, include it here
from corgie.layers.volumetric_layers import (
    FieldLayer,
    ImgLayer,
    MaskLayer,
    SectionValueLayer,
    SegmentationLayer,
    FixedFieldLayer,
    FloatTensorLayer,
)

# also update the default layer type if you want to
DEFAULT_LAYER_TYPE = "img"
